# This file is part of MyPaint.
# Copyright (C) 2013-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or

"""Combined menubar and toolbars."""


## Imports

from __future__ import division, print_function
import logging

from lib.gibindings import GObject
from lib.gibindings import Gtk
from lib.gibindings import Gdk
from gettext import gettext as _

logger = logging.getLogger(__name__)


## Class definitions

class TopBar (Gtk.Grid):
    """Combined menubar and toolbars which compacts when fullscreened.

    This is a container widget for two horizontal toolbars and a menubar
    with specialized behaviour when its parent window is fullscreened:
    the menubar is repacked into the toolbar, and temporary CSS styles
    are applied in order to attempt greater Fitts's Law compliance (and
    a nicer look).

    The toolbars and menubar are presented as properties for greater
    flexibility in construction. All of these properties must be set up
    at the time the widget is realized.

    """

    ## Class constants

    __gtype_name__ = 'MyPaintTopBar'
    ICON_NAMES = {
        "FileMenu": "mypaint-file-symbolic",
        "EditMenu": "mypaint-edit-symbolic",
        "ViewMenu": "mypaint-view-symbolic",
        "BrushMenu": "mypaint-brush-symbolic",
        "ColorMenu": "mypaint-colors-symbolic",
        "LayerMenu": "mypaint-layers-symbolic",
        "ScratchMenu": "mypaint-scratchpad-symbolic",
        "HelpMenu": "mypaint-help-symbolic",
    }

    ## GObject properties, for Builder-style construction

    #: The toolbar to present in position 1.
    toolbar1 = GObject.Property(
        type=Gtk.Toolbar,
        flags=GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        nick='Toolbar-1 widget',
        blurb="First GtkToolbar to show. This must be set at realize time."
    )

    #: The toolbar to present in position 2.
    toolbar1 = GObject.Property(
        type=Gtk.Toolbar,
        flags=GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        nick='Toolbar-2 widget',
        blurb="Second GtkToolbar to show. This must be set at realize time."
    )

    #: The menubar to present.
    menubar = GObject.Property(
        type=Gtk.MenuBar,
        flags=GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        nick='Menu Bar widget',
        blurb="The GtkMenuBar to show. This must be set at realize time."
    )

    ## Construction & initialization

    def __init__(self):
        """Initialize"""
        Gtk.Grid.__init__(self)
        self.connect("realize", self._realize_cb)
        # Widgets used in fullscreen mode
        fs_menu = Gtk.Menu()
        self._fs_menu = fs_menu
        self._fs_menubutton = FakeMenuButton(_("<b>MyPaint</b>"), fs_menu)
        self._fs_toolitem = Gtk.ToolItem()

    def _realize_cb(self, widget):
        """Assorted setup when the widget is realized"""
        assert self.menubar is not None
        assert self.toolbar1 is not None
        assert self.toolbar2 is not None
        # Packing details for Grid
        self.menubar.set_hexpand(True)
        self.toolbar1.set_hexpand(True)
        self.toolbar2.set_hexpand(False)
        self._fs_menubutton.set_hexpand(False)
        # Specialized styles
        prov = Gtk.CssProvider()
        prov.load_from_data(b"""
                .topbar {
                    padding: 0px; /* required by toolbars */
                    margin: 0px;  /* required by menubar */
                }
            """)
        bars = [self.toolbar1, self.toolbar2, self.menubar]
        for b in bars:
            style = b.get_style_context()
            style.add_provider(prov, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
            style.add_class("topbar")
        # Initial packing; assume a non-fullscreened state
        self.attach(self.menubar, 0, 0, 2, 1)
        self.attach(self.toolbar1, 0, 1, 1, 1)
        self.attach(self.toolbar2, 1, 1, 1, 1)
        # Track state transitions of the window's toplevel
        toplevel = self.get_toplevel()
        assert toplevel is not None
        toplevel.connect("window-state-event", self._toplevel_state_event_cb)

    ## Event handling

    def _toplevel_state_event_cb(self, toplevel, event):
        """Repacks widgets when the toplevel changes fullsceen state"""
        if not event.changed_mask & Gdk.WindowState.FULLSCREEN:
            return
        menubar = self.menubar
        toolbar1 = self.toolbar1
        toolbar2 = self.toolbar2
        assert self is toolbar1.get_parent()
        assert self is toolbar2.get_parent()
        if event.new_window_state & Gdk.WindowState.FULLSCREEN:
            # Remove menubar, use menu button on the 1st toolbar instead
            assert menubar.get_parent() is self
            assert self._fs_menubutton.get_parent() is None
            menubar.hide()
            self.remove(menubar)
            for menuitem in list(menubar):
                menubar.remove(menuitem)
                self._fs_menu.append(menuitem)
                item_name = menuitem.get_name()
                icon_name = self.ICON_NAMES.get(item_name, None)
                if hasattr(menuitem, "set_image"):
                    if icon_name:
                        icon_image = Gtk.Image()
                        icon_image.set_from_icon_name(
                            icon_name,
                            Gtk.IconSize.MENU,
                        )
                        menuitem.set_image(icon_image)
                    else:
                        logger.warning(
                            "No icon for %r in the fullscreen state",
                            item_name,
                        )
                menuitem.show_all()
            toolbar1.hide()
            self.remove(toolbar1)
            self.remove(toolbar2)
            self.attach(toolbar1, 0, 0, 1, 1)
            self.attach(toolbar2, 1, 0, 1, 1)
            toolbar1.insert(self._fs_toolitem, 0)
            self._fs_toolitem.add(self._fs_menubutton)
            self._fs_toolitem.show_all()
        else:
            # Windowed mode: use a regular menu bar above the toolbar
            assert menubar.get_parent() is None
            assert self._fs_menubutton.get_parent() is self._fs_toolitem
            toolbar1.remove(self._fs_toolitem)
            self._fs_toolitem.remove(self._fs_menubutton)
            for menuitem in list(self._fs_menu):
                self._fs_menu.remove(menuitem)
                menubar.append(menuitem)
                if hasattr(menuitem, "set_image"):
                    menuitem.set_image(None)
            toolbar1.hide()
            toolbar2.hide()
            self.remove(toolbar1)
            self.remove(toolbar2)
            self.attach(menubar, 0, 0, 2, 1)
            self.attach(toolbar1, 0, 1, 1, 1)
            self.attach(toolbar2, 1, 1, 1, 1)
            menubar.show()
        toolbar1.show_all()
        toolbar2.show_all()


class FakeMenuButton (Gtk.EventBox):
    """Button-styled widget that launches a dropdown menu when clicked"""

    def __init__(self, markup, menu):
        """Initialize

        :param markup: Markup to display in the button.
        :param menu: The menu to present when clicked.
        """
        Gtk.EventBox.__init__(self)
        self.menu = menu
        self.label = Gtk.Label()
        self.label.set_markup(markup)
        self.label.set_padding(8, 0)
        # Intercept mouse clicks and use them for activating the togglebutton
        # even if they're in its border, or (0, 0). Fitts would approve.
        invis = Gtk.EventBox()
        invis.set_visible_window(False)
        invis.set_above_child(True)
        invis.connect("button-press-event", self._button_press_cb)
        invis.connect("enter-notify-event", self._enter_cb)
        invis.connect("leave-notify-event", self._leave_cb)
        self.invis_window = invis
        # Toggle button, for the look of the thing only
        self.togglebutton = Gtk.ToggleButton()
        self.togglebutton.add(self.label)
        self.togglebutton.set_relief(Gtk.ReliefStyle.NONE)
        self.togglebutton.connect("toggled", self._togglebutton_toggled_cb)
        # The underlying togglebutton can default and focus. Might as well make
        # the Return key do something useful rather than invoking the 1st
        # toolbar item.
        self.togglebutton.set_can_default(True)
        self.togglebutton.set_can_focus(True)
        # Packing
        invis.add(self.togglebutton)
        self.add(invis)
        # Menu signals
        for sig in "selection-done", "deactivate", "cancel":
            menu.connect(sig, self._menu_dismiss_cb)

    def _enter_cb(self, widget, event):
        """Prelight the button when hovered"""
        self.togglebutton.set_state_flags(Gtk.StateFlags.PRELIGHT, False)

    def _leave_cb(self, widget, event):
        """Un-prelight the button when the pointer leaves"""
        self.togglebutton.unset_state_flags(Gtk.StateFlags.PRELIGHT)

    def _button_press_cb(self, widget, event):
        """Post the menmu when clicked

        Menu operation is much more convincing if we call popup() with event
        details here rather than leaving it to the child button's "toggled"
        event handler.
        """
        pos_func = self._get_popup_menu_position
        self.menu.popup(parent_menu_shell=None, parent_menu_item=None,
                        func=pos_func, data=None, button=event.button,
                        activate_time=event.time)
        self.togglebutton.set_active(True)

    def _togglebutton_toggled_cb(self, togglebutton):
        """Post the menu from a keypress activating the toggle

        The menu dismiss handler untoggles it."""
        if togglebutton.get_active():
            if not self.menu.get_property("visible"):
                pos_func = self._get_popup_menu_position
                self.menu.popup(None, None, pos_func, 1, 0)

    def _menu_dismiss_cb(self, *a, **kw):
        """Reset the button state when the user's finished

        Also transfer focus back to the menu button."""
        self.unset_state_flags(Gtk.StateFlags.PRELIGHT)
        self.togglebutton.set_active(False)
        self.togglebutton.grab_focus()

    def _get_popup_menu_position(self, menu, *junk):
        """Position function for menu popup

        This places the menu underneath the button, at the same x position.
        """
        win = self.get_window()
        x, y = win.get_origin()[1:]
        y += self.get_allocated_height()
        return x, y, True


## Testing

def _test():
    """Run an interactive test"""
    toplevel = Gtk.Window()
    toplevel.set_title("topbar test")
    toplevel.connect("destroy", lambda *a: Gtk.main_quit())
    mainbox = Gtk.VBox()
    topbar = TopBar()
    canvas = Gtk.DrawingArea()
    toplevel.set_size_request(500, 300)

    # Fullscreen action
    fs_act = Gtk.ToggleAction.new("Fullscreen", "Fullscreen",
                                  "Enter fullscreen mode",
                                  Gtk.STOCK_FULLSCREEN)

    def _fullscreen_cb(action, toplevel):
        if action.get_active():
            toplevel.fullscreen()
        else:
            toplevel.unfullscreen()
    fs_act.connect("toggled", _fullscreen_cb, toplevel)

    # One normally constructed menubar
    menu1 = Gtk.Menu()
    menuitem1 = Gtk.MenuItem.new_with_label("Demo")
    menuitem2 = Gtk.CheckMenuItem()
    menuitem3 = Gtk.MenuItem.new_with_label("Quit")
    menuitem1.set_submenu(menu1)
    menuitem2.set_related_action(fs_act)
    menuitem3.connect("activate", lambda *a: Gtk.main_quit())
    menu1.append(menuitem2)
    menu1.append(menuitem3)
    menuitem1.show()
    menuitem2.show()
    menuitem3.show()
    menubar = Gtk.MenuBar()
    menubar.append(menuitem1)

    # We need a pair of toolbars too.
    toolbar1 = Gtk.Toolbar()
    toolbar2 = Gtk.Toolbar()
    toolitem1 = Gtk.ToggleToolButton()
    toolitem1.set_related_action(fs_act)
    toolbar2.insert(toolitem1, -1)
    toolitem2 = Gtk.SeparatorToolItem()
    toolitem2.set_draw(False)
    toolitem2.set_expand(True)
    toolbar1.insert(toolitem2, 0)
    # Some junk items, to verify appearance in various GTK3 themes
    toolitem3 = Gtk.ToolButton.new_from_stock(Gtk.STOCK_ZOOM_100)
    toolitem3.set_is_important(True)
    toolitem4 = Gtk.ToolButton.new_from_stock(Gtk.STOCK_ZOOM_100)
    toolitem4.set_is_important(True)
    toolitem4.set_sensitive(False)
    toolitem5 = Gtk.ToolButton.new_from_stock(Gtk.STOCK_ZOOM_100)
    toolitem6 = Gtk.ToolButton.new_from_stock(Gtk.STOCK_ZOOM_100)
    toolitem6.set_sensitive(False)
    toolbar1.insert(toolitem6, 0)
    toolbar1.insert(toolitem5, 0)
    toolbar1.insert(toolitem4, 0)
    toolbar2.insert(toolitem3, 0)

    # Assign topbar's properties
    topbar.toolbar1 = toolbar1
    topbar.toolbar2 = toolbar2
    topbar.menubar = menubar

    # Pack main UI, and start demo
    mainbox.pack_start(topbar, False, False, 0)
    mainbox.pack_start(canvas, True, True, 0)
    toplevel.add(mainbox)
    toplevel.show_all()
    topbar.show_all()
    Gtk.main()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    _test()



# This file is part of MyPaint.
# Copyright (C) 2014 by Andrew Chadwick <a.t.chadwick@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or


"""Modes for manipulating the view"""


## Imports
from __future__ import division, print_function

import gui.mode

import math

from gettext import gettext as _


## Class defs

class PanViewMode (gui.mode.ScrollableModeMixin, gui.mode.OneshotDragMode):
    """A mode for translating the viewport by dragging."""

    ACTION_NAME = 'PanViewMode'

    pointer_behavior = gui.mode.Behavior.CHANGE_VIEW
    scroll_behavior = gui.mode.Behavior.CHANGE_VIEW
    supports_button_switching = True

    permitted_switch_actions = set([
        'RotateViewMode',
        'ZoomViewMode'
    ] + gui.mode.BUTTON_BINDING_ACTIONS)

    @classmethod
    def get_name(cls):
        return _(u"Scroll View")

    def get_usage(self):
        return _(u"Drag the canvas view")

    @property
    def inactive_cursor(self):
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME)

    @property
    def active_cursor(self):
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME)

    def drag_update_cb(self, tdw, event, dx, dy):
        # NOTE: keep in line with gui.mode.ScrollableViewMixin
        tdw.scroll(-dx, -dy)
        self.doc.notify_view_changed()
        super(PanViewMode, self).drag_update_cb(tdw, event, dx, dy)


class ZoomViewMode (gui.mode.ScrollableModeMixin, gui.mode.OneshotDragMode):
    """A mode for zooming the viewport by dragging."""

    ACTION_NAME = 'ZoomViewMode'

    pointer_behavior = gui.mode.Behavior.CHANGE_VIEW
    scroll_behavior = gui.mode.Behavior.CHANGE_VIEW
    supports_button_switching = True

    permitted_switch_actions = set([
        'PanViewMode'
        'RotateViewMode',
    ] + gui.mode.BUTTON_BINDING_ACTIONS)

    @classmethod
    def get_name(cls):
        return _(u"Zoom View")

    def get_usage(self):
        return _(u"Zoom the canvas view")

    @property
    def active_cursor(self):
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME)

    @property
    def inactive_cursor(self):
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME)

    def drag_update_cb(self, tdw, event, dx, dy):
        tdw.scroll(-dx, -dy)
        tdw.zoom(math.exp(dy/100.0), center=(event.x, event.y))
        # TODO: Let modifiers constrain the zoom amount to
        # TODO: the defined steps. Shift seems to be conventional now...
        self.doc.notify_view_changed()
        super(ZoomViewMode, self).drag_update_cb(tdw, event, dx, dy)
        # TODO: The constrain amount and threshold should be similar to
        # TODO: what gui.mode.ScrollableModeMixin now does.


class RotateViewMode (gui.mode.ScrollableModeMixin, gui.mode.OneshotDragMode):
    """A mode for rotating the viewport by dragging."""

    ACTION_NAME = 'RotateViewMode'

    pointer_behavior = gui.mode.Behavior.CHANGE_VIEW
    scroll_behavior = gui.mode.Behavior.CHANGE_VIEW
    supports_button_switching = True

    permitted_switch_actions = set([
        'PanViewMode'
        'ZoomViewMode',
    ] + gui.mode.BUTTON_BINDING_ACTIONS)

    @classmethod
    def get_name(cls):
        return _(u"Rotate View")

    def get_usage(cls):
        return _(u"Rotate the canvas view")

    @property
    def active_cursor(self):
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME)

    @property
    def inactive_cursor(self):
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME)

    def drag_update_cb(self, tdw, event, dx, dy):
        # calculate angular velocity from the rotation center
        x, y = event.x, event.y
        cx, cy = tdw.get_center()
        x, y = x-cx, y-cy
        phi2 = math.atan2(y, x)
        x, y = x-dx, y-dy
        phi1 = math.atan2(y, x)
        tdw.rotate(phi2-phi1, center=(cx, cy))
        self.doc.notify_view_changed()
        # TODO: Allow modifiers to constrain the transformation angle
        #       to 22.5 degree steps.
        super(RotateViewMode, self).drag_update_cb(tdw, event, dx, dy)



# This file is part of MyPaint.
# Copyright (C) 2009-2018 by the MyPaint Development Team.
# Copyright (C) 2007-2012 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""This module implements an unbounded tiled surface for painting."""

## Imports

from __future__ import division, print_function
import time
import sys
import os
import contextlib
import logging

from gettext import gettext as _
import numpy as np

from . import mypaintlib
from . import pixbufsurface
from lib.eotf import eotf
import lib.surface
from lib.surface import TileAccessible
from lib.surface import TileBlittable
from lib.surface import TileCompositable
from .errors import FileHandlingError
import lib.fileutils
import lib.modes
import lib.feedback
import lib.floodfill
from lib.pycompat import xrange
from lib.pycompat import PY3, itervalues

logger = logging.getLogger(__name__)

## Constants

TILE_SIZE = N = mypaintlib.TILE_SIZE
MAX_MIPMAP_LEVEL = mypaintlib.MAX_MIPMAP_LEVEL

SYMMETRY_TYPES = tuple(range(mypaintlib.NumSymmetryTypes))
SYMMETRY_STRINGS = {
    mypaintlib.SymmetryVertical: _("Vertical"),
    mypaintlib.SymmetryHorizontal: _("Horizontal"),
    mypaintlib.SymmetryVertHorz: _("Vertical and horizontal"),
    mypaintlib.SymmetryRotational: _("Rotational"),
    mypaintlib.SymmetrySnowflake: _("Snowflake"),
}
for sym_type in SYMMETRY_TYPES:
    assert sym_type in SYMMETRY_STRINGS


## Tile class and marker tile constants

class _Tile (object):
    """Internal tile storage, with readonly flag

    Note: pixels are stored with premultiplied alpha.
    15 bits are used, but fully opaque or white is stored as 2**15
    (requiring 16 bits). This is to allow many calculations to divide by
    2**15 instead of (2**16-1).

    """

    def __init__(self, copy_from=None):
        super(_Tile, self).__init__()
        if copy_from is None:
            self.rgba = np.zeros((N, N, 4), 'uint16')
        else:
            self.rgba = copy_from.rgba.copy()
        self.readonly = False

    def copy(self):
        return _Tile(copy_from=self)


# tile for read-only operations on empty spots
transparent_tile = _Tile()
transparent_tile.readonly = True

# tile with invalid pixel memory (needs refresh)
mipmap_dirty_tile = _Tile()
del mipmap_dirty_tile.rgba


## Class defs: surfaces

class _SurfaceSnapshot (object):
    pass


# TODO:
# - move the tile storage from MyPaintSurface to a separate class
class MyPaintSurface (TileAccessible, TileBlittable, TileCompositable):
    """Tile-based surface

    The C++ part of this class is in tiledsurface.hpp
    """

    def __init__(self, mipmap_level=0, mipmap_surfaces=None,
                 looped=False, looped_size=(0, 0)):
        super(MyPaintSurface, self).__init__()

        # TODO: pass just what it needs access to, not all of self
        self._backend = mypaintlib.TiledSurface(self)
        self.tiledict = {}
        self.observers = []

        # Used to implement repeating surfaces, like Background
        if looped_size[0] % N or looped_size[1] % N:
            raise ValueError('Looped size must be multiples of tile size')
        self.looped = looped
        self.looped_size = looped_size

        self.mipmap_level = mipmap_level
        if mipmap_level == 0:
            assert mipmap_surfaces is None
            self._mipmaps = self._create_mipmap_surfaces()
        else:
            assert mipmap_surfaces is not None
            self._mipmaps = mipmap_surfaces

        # Forwarding API
        self.set_symmetry_state = self._backend.set_symmetry_state
        self.begin_atomic = self._backend.begin_atomic

        self.get_color = self._backend.get_color
        self.get_alpha = self._backend.get_alpha
        self.draw_dab = self._backend.draw_dab

    @classmethod
    def _mock(cls):
        """Returns a new mock surface with a pattern."""
        surf = cls()
        v1 = (1 << 15)
        v0 = 0
        black = (v0, v0, v0, v1)
        white = (v1, v1, v1, v1)
        bits = [(1, 0), (0, 1), (1, 1), (2, 1), (0, 2), (2, 2)]
        b = 1
        for tx in range(b+b+1+max(xt for (xt, yt) in bits)):
            for ty in range(b+b+1+max(yt for (xt, yt) in bits)):
                p = ((tx-b, ty-b) in bits) and white or black
                with surf.tile_request(tx, ty, readonly=False) as rgba:
                    rgba[:] = p
        return surf

    def _create_mipmap_surfaces(self):
        """Internal: initializes an internal mipmap lookup table

        Overridable to avoid unnecessary work when initializing the background
        surface subclass.
        """
        assert self.mipmap_level == 0
        mipmaps = [self]
        for level in range(1, MAX_MIPMAP_LEVEL+1):
            s = MyPaintSurface(mipmap_level=level, mipmap_surfaces=mipmaps)
            mipmaps.append(s)

        # for quick lookup
        for level, s in enumerate(mipmaps):
            try:
                s.parent = mipmaps[level-1]
            except IndexError:
                s.parent = None
            try:
                s.mipmap = mipmaps[level+1]
            except IndexError:
                s.mipmap = None
        return mipmaps

    def end_atomic(self):
        bboxes = self._backend.end_atomic()
        for bbox in bboxes:
            if (bbox[2] > 0 and bbox[3] > 0):
                self.notify_observers(*bbox)

    @property
    def backend(self):
        return self._backend

    def notify_observers(self, *args):
        for f in self.observers:
            f(*args)

    def clear(self):
        tiles = self.tiledict.keys()
        self.tiledict = {}
        self.notify_observers(*lib.surface.get_tiles_bbox(tiles))
        if self.mipmap:
            self.mipmap.clear()

    def trim(self, rect):
        """Trim the layer to a rectangle, discarding data outside it

        :param rect: A trimming rectangle in model coordinates
        :type rect: tuple (x, y, w, h)

        Only complete tiles are discarded by this method.
        If a tile is neither fully inside nor fully outside the
        rectangle, the part of the tile outside the rectangle will be
        cleared.
        """
        x, y, w, h = rect
        logger.info("Trim %dx%d%+d%+d", w, h, x, y)
        trimmed = []
        for tx, ty in list(self.tiledict.keys()):
            if tx*N+N < x or ty*N+N < y or tx*N > x+w or ty*N > y+h:
                trimmed.append((tx, ty))
                self.tiledict.pop((tx, ty))
                self._mark_mipmap_dirty(tx, ty)
            elif (tx*N < x and x < tx*N+N
                    or ty*N < y and y < ty*N+N
                    or tx*N < x+w and x+w < tx*N+N
                    or ty*N < y+h and y+h < ty*N+N):
                trimmed.append((tx, ty))
                with self.tile_request(tx, ty, readonly=False) as rgba:
                    if tx*N < x and x < tx*N+N:
                        rgba[:, 0:(x - tx*N), :] = 0  # Clear left edge

                    if ty*N < y and y < ty*N+N:
                        rgba[0:(y - ty*N), :, :] = 0  # Clear top edge

                    if tx*N < x+w and x+w < tx*N+N:
                        # This slice is [N-1-c for c in range(tx*N+N - (x+w))].
                        rgba[:, (x+w - tx*N):N, :] = 0  # Clear right edge

                    if ty*N < y+h and y+h < ty*N+N:
                        # This slice is [N-1-r for r in range(ty*N+N - (y+h))].
                        rgba[(y+h - ty*N):N, :, :] = 0  # Clear bottom edge
                self._mark_mipmap_dirty(tx, ty)

        self.notify_observers(*lib.surface.get_tiles_bbox(trimmed))

    @contextlib.contextmanager
    def tile_request(self, tx, ty, readonly):
        """Get a tile as a NumPy array, then put it back

        :param int tx: Tile X coord (multiply by TILE_SIZE for pixels)
        :param int ty: Tile Y coord (multiply by TILE_SIZE for pixels)
        :param bool readonly: get a read-only tile

        Context manager that fetches a tile as a NumPy array,
        and then puts the potentially modified tile back into the
        tile backing store. To be used with the 'with' statement.
        Read/write tile requests on empty slots get you a new
        writeable tile::

            >>> surf = MyPaintSurface()
            >>> with surf.tile_request(1, 2, readonly=False) as t1:
            ...     t1[...] = (1<<15)

            >>> with surf.tile_request(1, 2, readonly=False) as t2:
            ...     assert t2 is t1
            ...     assert (t2 == t1).all()

        Read-only tile requests on empty addresses yield the special
        transparent tile, which is marked as read-only::

            >>> with surf.tile_request(666, 666, readonly=True) as tr:
            ...     assert tr is transparent_tile.rgba

        Snapshotting a surface makes all its tiles read-only as a side
        effect, so the next read/write tile request will yield a copy
        for you to work on::

            >>> sshot = surf.save_snapshot()
            >>> with surf.tile_request(1, 2, readonly=True) as t3:
            ...     assert t3 is t1
            ...     assert (t3 == t1).all()
            >>> with surf.tile_request(1, 2, readonly=False) as t4:
            ...     assert t4 is not t1
            ...     assert (t4 == t1).all()

        """
        numpy_tile = self._get_tile_numpy(tx, ty, readonly)
        yield numpy_tile
        self._set_tile_numpy(tx, ty, numpy_tile, readonly)

    def _regenerate_mipmap(self, t, tx, ty):
        t = _Tile()
        self.tiledict[(tx, ty)] = t
        empty = True

        for x in xrange(2):
            for y in xrange(2):
                src = self.parent.tiledict.get((tx*2 + x, ty*2 + y),
                                               transparent_tile)
                if src is mipmap_dirty_tile:
                    src = self.parent._regenerate_mipmap(
                        src,
                        tx*2 + x, ty*2 + y,
                    )
                mypaintlib.tile_downscale_rgba16(src.rgba, t.rgba,
                                                 x * N // 2,
                                                 y * N // 2)
                if src.rgba is not transparent_tile.rgba:
                    empty = False
        if empty:
            # rare case, no need to speed it up
            del self.tiledict[(tx, ty)]
            t = transparent_tile
        return t

    def _get_tile_numpy(self, tx, ty, readonly):
        # OPTIMIZE: do some profiling to check if this function is a bottleneck
        #           yes it is
        # Note: we must return memory that stays valid for writing until the
        # last end_atomic(), because of the caching in tiledsurface.hpp.

        if self.looped:
            tx = tx % (self.looped_size[0] // N)
            ty = ty % (self.looped_size[1] // N)

        t = self.tiledict.get((tx, ty))
        if t is None:
            if readonly:
                t = transparent_tile
            else:
                t = _Tile()
                self.tiledict[(tx, ty)] = t
        if t is mipmap_dirty_tile:
            t = self._regenerate_mipmap(t, tx, ty)
        if t.readonly and not readonly:
            # shared memory, get a private copy for writing
            t = t.copy()
            self.tiledict[(tx, ty)] = t
        if not readonly:
            # assert self.mipmap_level == 0
            self._mark_mipmap_dirty(tx, ty)
        return t.rgba

    def _set_tile_numpy(self, tx, ty, obj, readonly):
        pass  # Data can be modified directly, no action needed

    def _mark_mipmap_dirty(self, tx, ty):
        # assert self.mipmap_level == 0
        if not self._mipmaps:
            return
        for level, mipmap in enumerate(self._mipmaps):
            if level == 0:
                continue
            fac = 2**(level)
            if mipmap.tiledict.get((tx // fac, ty // fac),
                                   None) == mipmap_dirty_tile:
                break
            mipmap.tiledict[(tx // fac, ty // fac)] = mipmap_dirty_tile

    def blit_tile_into(self, dst, dst_has_alpha, tx, ty, mipmap_level=0,
                       *args, **kwargs):
        """Copy one tile from this object into a destination array

        See lib.surface.TileBlittable for the parameters. This
        implementation adds an extra param:

        :param int mipmap_level: layer mipmap level to use

        """

        # assert dst_has_alpha is True

        if self.mipmap_level < mipmap_level:
            return self.mipmap.blit_tile_into(dst, dst_has_alpha, tx, ty,
                                              mipmap_level)

        assert dst.shape[2] == 4
        if dst.dtype not in ('uint16', 'uint8'):
            raise ValueError('Unsupported destination buffer type %r',
                             dst.dtype)
        dst_is_uint16 = (dst.dtype == 'uint16')

        with self.tile_request(tx, ty, readonly=True) as src:
            if src is transparent_tile.rgba:
                # dst[:] = 0  # <-- notably slower than memset()
                if dst_is_uint16:
                    mypaintlib.tile_clear_rgba16(dst)
                else:
                    mypaintlib.tile_clear_rgba8(dst)
            else:
                if dst_is_uint16:
                    # this will do memcpy, not worth to bother skipping
                    # the u channel
                    mypaintlib.tile_copy_rgba16_into_rgba16(src, dst)
                else:
                    if dst_has_alpha:
                        mypaintlib.tile_convert_rgba16_to_rgba8(
                            src, dst, eotf()
                        )
                    else:
                        mypaintlib.tile_convert_rgbu16_to_rgbu8(
                            src, dst, eotf()
                        )

    def composite_tile(self, dst, dst_has_alpha, tx, ty, mipmap_level=0,
                       opacity=1.0, mode=mypaintlib.CombineNormal,
                       *args, **kwargs):
        """Composite one tile of this surface over a NumPy array.

        See lib.surface.TileCompositable for the parameters. This
        implementation adds two further ones:

        :param float opacity: opacity multiplier
        :param int mode: mode to use when compositing

        """

        # Apply zero-alpha-source optimizations if possible.
        # Sometimes this can be done without issuing a tile request.
        if opacity == 0:
            if dst_has_alpha:
                if mode in lib.modes.MODES_CLEARING_BACKDROP_AT_ZERO_ALPHA:
                    mypaintlib.tile_clear_rgba16(dst)
                    return
            if mode not in lib.modes.MODES_EFFECTIVE_AT_ZERO_ALPHA:
                return

        # Tile request needed, but may need to satisfy it from a deeper
        # mipmap level.
        if self.mipmap_level < mipmap_level:
            self.mipmap.composite_tile(dst, dst_has_alpha, tx, ty,
                                       mipmap_level, opacity, mode)
            return

        # Tile request at the required level.
        # Try optimizations again if we got the special marker tile
        with self.tile_request(tx, ty, readonly=True) as src:
            if src is transparent_tile.rgba:
                if dst_has_alpha:
                    if mode in lib.modes.MODES_CLEARING_BACKDROP_AT_ZERO_ALPHA:
                        mypaintlib.tile_clear_rgba16(dst)
                        return
                if mode not in lib.modes.MODES_EFFECTIVE_AT_ZERO_ALPHA:
                    return
            mypaintlib.tile_combine(mode, src, dst, dst_has_alpha, opacity)

    ## Snapshotting

    def save_snapshot(self):
        """Creates and returns a snapshot of the surface

        Snapshotting marks all the tiles of the surface as read-only,
        then just shallow-copes the tiledict. It's quick. See
        tile_request() for how new read/write tiles can be unlocked.

        """
        sshot = _SurfaceSnapshot()
        for t in itervalues(self.tiledict):
            t.readonly = True
        sshot.tiledict = self.tiledict.copy()
        return sshot

    def load_snapshot(self, sshot):
        """Loads a saved snapshot, replacing the internal tiledict"""
        self._load_tiledict(sshot.tiledict)

    def _load_tiledict(self, d):
        """Efficiently loads a tiledict, and notifies the observers"""
        if d == self.tiledict:
            # common case optimization, called via stroke.redo()
            # testcase: comparison above (if equal) takes 0.6ms,
            # code below 30ms
            return
        old = set(self.tiledict.items())
        self.tiledict = d.copy()
        new = set(self.tiledict.items())
        dirty = old.symmetric_difference(new)
        for pos, tile in dirty:
            self._mark_mipmap_dirty(*pos)
        bbox = lib.surface.get_tiles_bbox(pos for (pos, tile) in dirty)
        if not bbox.empty():
            self.notify_observers(*bbox)

    ## Loading tile data

    def load_from_surface(self, other):
        """Loads tile data from another surface, via a snapshot"""
        self.load_snapshot(other.save_snapshot())

    def _load_from_pixbufsurface(self, s):
        dirty_tiles = set(self.tiledict.keys())
        self.tiledict = {}

        for tx, ty in s.get_tiles():
            with self.tile_request(tx, ty, readonly=False) as dst:
                s.blit_tile_into(dst, True, tx, ty)

        dirty_tiles.update(self.tiledict.keys())
        bbox = lib.surface.get_tiles_bbox(dirty_tiles)
        self.notify_observers(*bbox)

    def load_from_numpy(self, arr, x, y):
        """Loads tile data from a numpy array

        :param arr: Array containing the pixel data
        :type arr: numpy.ndarray of uint8, dimensions HxWx3 or HxWx4
        :param x: X coordinate for the array
        :param y: Y coordinate for the array
        :returns: the dimensions of the loaded surface, as (x,y,w,h)

        """
        h, w, channels = arr.shape
        if h <= 0 or w <= 0:
            return (x, y, w, h)

        if arr.dtype == 'uint8':
            s = pixbufsurface.Surface(x, y, w, h, data=arr)
            self._load_from_pixbufsurface(s)
        else:
            raise ValueError("Only uint8 data is supported by MyPaintSurface")

        return (x, y, w, h)

    def load_from_png(self, filename, x, y, progress=None,
                      convert_to_srgb=True,
                      **kwargs):
        """Load from a PNG, one tilerow at a time, discarding empty tiles.

        :param str filename: The file to load
        :param int x: X-coordinate at which to load the replacement data
        :param int y: Y-coordinate at which to load the replacement data
        :param bool convert_to_srgb: If True, convert to sRGB
        :param progress: Unsized UI feedback obj.
        :type progress: lib.feedback.Progress or None
        :param dict \*\*kwargs: Ignored

        Raises a `lib.errors.FileHandlingError` with a descriptive
        string when conversion or PNG reading fails.

        """
        if not progress:
            progress = lib.feedback.Progress()

        # Catch this error up front, since the code below hides it.
        if progress.items is not None:
            raise ValueError("progress arg must be unsized")

        dirty_tiles = set(self.tiledict.keys())
        self.tiledict = {}

        ty0 = int(y // N)
        state = {}
        state['buf'] = None   # array of height N, width depends on image
        state['ty'] = ty0  # current tile row being filled into buf
        state['frame_size'] = None
        state['progress'] = progress

        def get_buffer(png_w, png_h):
            if state["frame_size"] is None:
                if state['progress']:
                    ty_final = int((y + png_h) // N)
                    # We have to handle feedback exceptions ourself
                    try:
                        state["progress"].items = ty_final - ty0
                    except Exception:
                        logger.exception("setting progress.items failed")
                        state["progress"] = None
                state['frame_size'] = (x, y, png_w, png_h)

            buf_x0 = x // N * N
            buf_x1 = ((x + png_w - 1) // N + 1) * N
            buf_y0 = state['ty']*N
            buf_y1 = buf_y0+N
            buf_w = buf_x1-buf_x0
            buf_h = buf_y1-buf_y0
            assert buf_w % N == 0
            assert buf_h == N
            if state['buf'] is not None:
                consume_buf()
            else:
                state['buf'] = np.empty((buf_h, buf_w, 4), 'uint8')

            png_x0 = x
            png_x1 = x+png_w
            subbuf = state['buf'][:, png_x0-buf_x0:png_x1-buf_x0]
            if 1:  # optimize: only needed for first and last
                state['buf'].fill(0)
                png_y0 = max(buf_y0, y)
                png_y1 = min(buf_y0+buf_h, y+png_h)
                assert png_y1 > png_y0
                subbuf = subbuf[png_y0-buf_y0:png_y1-buf_y0, :]

            state['ty'] += 1
            return subbuf

        def consume_buf():
            ty = state['ty']-1
            for i in xrange(state['buf'].shape[1] // N):
                tx = x // N + i
                src = state['buf'][:, i*N:(i+1)*N, :]
                if src[:, :, 3].any():
                    with self.tile_request(tx, ty, readonly=False) as dst:
                        mypaintlib.tile_convert_rgba8_to_rgba16(
                            src, dst, eotf()
                        )
            if state["progress"]:
                try:
                    state["progress"].completed(ty - ty0)
                except Exception:
                    logger.exception("Progress.completed() failed")
                    state["progress"] = None

        if sys.platform == 'win32':
            filename_sys = filename.encode("utf-8")
        else:
            filename_sys = filename.encode(sys.getfilesystemencoding())
            # FIXME: should not do that, should use open(unicode_object)

        if PY3:
            filename_sys = filename_sys.decode()
            # FIXME: https://github.com/mypaint/mypaint/issues/906

        try:
            flags = mypaintlib.load_png_fast_progressive(
                filename_sys,
                get_buffer,
                convert_to_srgb,
            )
        except (IOError, OSError, RuntimeError) as ex:
            raise FileHandlingError(_("PNG reader failed: %s") % str(ex))
        consume_buf()  # also process the final chunk of data
        progress.close()
        logger.debug("PNG loader flags: %r", flags)

        dirty_tiles.update(self.tiledict.keys())
        bbox = lib.surface.get_tiles_bbox(dirty_tiles)
        self.notify_observers(*bbox)

        # return the bbox of the loaded image
        return state['frame_size']

    def render_as_pixbuf(self, *args, **kwargs):
        if not self.tiledict:
            logger.warning('empty surface')
        t0 = time.time()
        kwargs['alpha'] = True
        res = pixbufsurface.render_as_pixbuf(self, *args, **kwargs)
        logger.debug('%.3fs rendering layer as pixbuf', time.time() - t0)
        return res

    def save_as_png(self, filename, *args, **kwargs):
        if 'alpha' not in kwargs:
            kwargs['alpha'] = True

        if len(self.tiledict) == 1 and self.looped:
            kwargs['single_tile_pattern'] = True
        lib.surface.save_as_png(self, filename, *args, **kwargs)

    def get_bbox(self):
        return lib.surface.get_tiles_bbox(self.tiledict)

    def get_tiles(self):
        return self.tiledict

    def is_empty(self):
        return not self.tiledict

    def remove_empty_tiles(self):
        """Removes tiles from the tiledict which contain no data"""
        if self.mipmap_level != 0:
            raise ValueError("Only call this on the top-level surface.")
        assert self is self._mipmaps[0]
        total = 0
        removed = 0
        for surf in self._mipmaps:
            tmp_items_list = surf.tiledict.items()
            if PY3:
                tmp_items_list = list(tmp_items_list)
            for pos, data in tmp_items_list:
                total += 1
                try:
                    rgba = data.rgba
                except AttributeError:
                    continue
                if rgba.any():
                    continue
                surf.tiledict.pop(pos)
                removed += 1
        return removed, total

    def remove_tiles(self, indices):
        """Removes a set of tiles from the surface by tile index."""
        if self.mipmap_level != 0:
            raise ValueError("Only call this on the top-level surface.")

        removed = set()
        for tx, ty in indices:
            pos = (tx, ty)
            if pos not in self.tiledict:
                continue
            self.tiledict.pop(pos)
            removed.add(pos)
            self._mark_mipmap_dirty(tx, ty)

        bbox = lib.surface.get_tiles_bbox(removed)
        self.notify_observers(*bbox)

    def get_move(self, x, y, sort=True):
        """Returns a move object for this surface

        :param x: Start position for the move, X coord
        :param y: Start position for the move, X coord
        :param sort: If true, sort tiles to move by distance from (x,y)
        :rtype: _TiledSurfaceMove

        It's up to the caller to ensure that only one move is active at a
        any single instant in time.
        """
        if self.mipmap_level != 0:
            raise ValueError("Only call this on the top-level surface.")
        return _TiledSurfaceMove(self, x, y, sort=sort)

    def flood_fill(self, fill_args, dst):
        """Fills connected areas of this surface into another

        :param fill_args: fill arguments object
        :type fill_args: lib.floodfill.FloodFillArguments
        :param dst: Target surface
        :type dst: MyPaintSurface
        """

        return flood_fill(self, fill_args, dst)

    @contextlib.contextmanager
    def cairo_request(self, x, y, w, h, mode=lib.modes.default_mode):
        """Get a Cairo context for a given area, then put back changes.

        :param int x: Request area's X coordinate.
        :param int y: Request area's Y coordinate.
        :param int w: Request area's width.
        :param int h: Request area's height.
        :param mode: Combine mode for the put-back.
        :rtype: contextlib.GeneratorContextManager
        :returns: cairo.Context (via with-statement)

        This context manager works by constructing and managing a
        temporary 8bpp Cairo imagesurface and yielding up a Cairo
        context that points at it. Call the method as part of a
        with-statement:

        >>> s = MyPaintSurface()
        >>> n = TILE_SIZE
        >>> assert n > 10
        >>> with s.cairo_request(10, 10, n, n) as cr:
        ...     cr.set_source_rgb(1, 0, 0)
        ...     cr.rectangle(1, 1, n-2, n-2)
        ...     cr.fill()
        >>> list(sorted(s.tiledict.keys()))
        [(0, 0), (0, 1), (1, 0), (1, 1)]

        If the mode is specified, it must be a layer/surface combine
        mode listed in `lib.modes.STACK_MODES`. In this case, The
        temporary cairo ImageSurface object is initially completely
        transparent, and anything you draw to it is composited back over
        the surface using the mode you requested.

        If you pass mode=None (or mode=lib.modes.PASS_THROUGH_MODE),
        Cairo operates directly on the surface. This means that you can
        use Cairo operators and primitives which erase tile data.

        >>> import cairo
        >>> with s.cairo_request(0, 0, n, n, mode=None) as cr:
        ...     cr.set_operator(cairo.OPERATOR_CLEAR)
        ...     cr.set_source_rgb(1, 1, 1)
        ...     cr.paint()
        >>> _ignored = s.remove_empty_tiles()
        >>> list(sorted(s.tiledict.keys()))
        [(0, 1), (1, 0), (1, 1)]
        >>> with s.cairo_request(n-10, n-10, n+10, n+10, mode=None) as cr:
        ...     cr.set_operator(cairo.OPERATOR_SOURCE)
        ...     cr.set_source_rgba(0, 1, 0, 0)
        ...     cr.paint()
        >>> _ignored = s.remove_empty_tiles()
        >>> list(sorted(s.tiledict.keys()))
        [(0, 1), (1, 0)]

        See also:

        * lib.pixbufsurface.Surface.cairo_request()
        * lib.layer.data.SimplePaintingMode.cairo_request()

        """

        # Normalize and validate args
        if callable(mode):
            mode = mode()

        if mode is not None:
            if mode == lib.modes.PASS_THROUGH_MODE:
                mode = None
            elif mode not in lib.modes.STANDARD_MODES:
                raise ValueError(
                    "The 'mode' argument must be one of STANDARD_MODES, "
                    "or it must be either PASS_THROUGH_MODE or None."
                )
        x = int(x)
        y = int(y)
        w = int(w)
        h = int(h)
        if w <= 0 or h <= 0:
            return   # nothing to do

        # Working pixbuf-surface
        s = lib.pixbufsurface.Surface(x, y, w, h)
        dirty_tiles = list(s.get_tiles())

        # Populate it with an 8-bit downsampling of this surface's data
        # if we're going to be "working directly"
        if mode is None:
            for tx, ty in dirty_tiles:
                with s.tile_request(tx, ty, readonly=False) as dst:
                    self.blit_tile_into(dst, True, tx, ty)

        # Collect changes from the invoker...
        with s.cairo_request() as cr:
            yield cr

        # Blit or composite back the changed pixbuf
        if mode is None:
            for tx, ty in dirty_tiles:
                with self.tile_request(tx, ty, readonly=False) as dst:
                    s.blit_tile_into(dst, True, tx, ty)
        else:
            tmp = np.zeros((N, N, 4), 'uint16')
            for tx, ty in dirty_tiles:
                s.blit_tile_into(tmp, True, tx, ty)
                with self.tile_request(tx, ty, readonly=False) as dst:
                    mypaintlib.tile_combine(mode, tmp, dst, True, 1.0)

        # Tell everyone about the changes
        bbox = lib.surface.get_tiles_bbox(dirty_tiles)
        self.notify_observers(*bbox)


class _TiledSurfaceMove (object):
    """Ongoing move state for a tiled surface, processed in chunks

    Tile move processing involves slicing and copying data from a
    snapshot of the surface's original tile arrays into an active
    surface within the model document. It's therefore potentially very
    slow for huge layers: doing this interactively requires the move to
    be processed in chunks in idle routines.

    Moves are created by a surface's get_move() method starting at a
    particular point in model coordinates.

        >>> surf = MyPaintSurface()
        >>> with surf.tile_request(10, 10, readonly=False) as a:
        ...     a[...] = 1<<15
        >>> len(surf.tiledict)
        1
        >>> move = surf.get_move(N/2, N/2, sort=True)

    During an interactive move, the move object is typically updated in
    response to the user moving the pointer,

        >>> move.update(N/2, N/2)
        >>> move.update(N/2 + 1, N/2 + 3)

    while being processed in chunks of a few hundred tiles in an idle
    routine.

        >>> while move.process():
        ...     pass

    When the user is done moving things and releases the layer, or quits
    the layer moving mode, the conventional way of finalizing things is

        >>> move.process(n=-1)
        False
        >>> move.cleanup()

    After the cleanup, the move should not be updated or processed any
    further.

    Moves which are not an exact multiple of the tile size generally
    make more tiles due to slicing and recombining.

        >>> len(surf.tiledict)
        4

    Moves which are an exact multiple of the tile size are processed
    faster (and never add tiles to the layer).

        >>> surf = MyPaintSurface()
        >>> with surf.tile_request(-3, 2, readonly=False) as a:
        ...     a[...] = 1<<15
        >>> list(surf.tiledict.keys())
        [(-3, 2)]
        >>> move = surf.get_move(0, 0, sort=False)
        >>> move.update(N*3, -N*2)
        >>> move.process(n=1)   # single op suffices
        False
        >>> move.cleanup()
        >>> list(surf.tiledict.keys())
        [(0, 0)]
        >>> # Please excuse the doctest for this special case
        >>> # just regression-proofing.

    Moves can be processed non-interactively by calling all the
    different phases together, as above.

    """

    def __init__(self, surface, x, y, sort=True):
        """Starts the move, recording state in the Move object

        :param x: Where to start, model X coordinate
        :param y: Where to start, model Y coordinate
        :param sort: If true, sort tiles to move by distance from (x,y)

        Sorting tiles by distance makes the move look nicer when moving
        interactively, but it's pointless for non-interactive moves.

        """
        object.__init__(self)
        self.surface = surface
        self.snapshot = surface.save_snapshot()
        self.chunks = []
        self.chunks[:] = self.snapshot.tiledict.keys()
        self.sort = sort
        tx = x // N
        ty = y // N
        self.start_pos = (x, y)
        if self.sort:
            self.chunks.sort(
                key=lambda p: abs(tx - p[0]) + abs(ty - p[1]),
            )
        # High water mark of chunks processed so far.
        # This is reset on every call to update().
        self.chunks_i = 0
        # Tile state tracking for individual update cycles
        self.written = set()
        self.blank_queue = []
        # Tile offsets which we'll be applying,
        # initially the move is zero.
        self.slices_x = calc_translation_slices(0)
        self.slices_y = calc_translation_slices(0)

    def update(self, dx, dy):
        """Updates the offset during a move

        :param dx: New move offset: relative to the constructor x.
        :param dy: New move offset: relative to the constructor y.

        This causes all the move's work to be re-queued.
        """
        # Nothing has been written in this pass yet
        self.written = set()
        # Tile indices to be cleared during processing,
        # unless they've been written to
        self.blank_queue[:] = self.surface.tiledict.keys()  # fresh!
        if self.sort:
            x, y = self.start_pos
            tx = (x + dx) // N
            ty = (y + dy) // N
            self.blank_queue.sort(
                key=lambda p: abs(tx - p[0]) + abs(ty - p[1]),
            )
        # Calculate offsets
        self.slices_x = calc_translation_slices(int(dx))
        self.slices_y = calc_translation_slices(int(dy))
        # Need to process every source chunk
        self.chunks_i = 0

    def cleanup(self):
        """Cleans up after processing the move.

        This must be called after the move has been processed fully, and
        should only be called after `process()` indicates that all tiles have
        been sliced and moved.

        """
        # Process any remaining work. Caller should have done this already.
        if self.chunks_i < len(self.chunks) or len(self.blank_queue) > 0:
            logger.warning("Stuff left to do at end of move cleanup(). May "
                           "result in poor interactive appearance. "
                           "chunks=%d/%d, blanks=%d", self.chunks_i,
                           len(self.chunks), len(self.blank_queue))
            logger.warning("Doing cleanup now...")
            self.process(n=-1)
        assert self.chunks_i >= len(self.chunks)
        assert len(self.blank_queue) == 0
        # Remove empty tiles created by Layer Move
        removed, total = self.surface.remove_empty_tiles()
        logger.debug(
            "_TiledSurfaceMove.cleanup: removed %d empty tiles of %d",
            removed, total,
        )

    def process(self, n=200):
        """Process a number of pending tile moves

        :param int n: The number of source tiles to process in this call
        :returns: whether there are any more tiles to process
        :rtype: bool

        Specify zero or negative `n` to process all remaining tiles.

        """
        updated = set()
        moves_remaining = self._process_moves(n, updated)
        blanks_remaining = self._process_blanks(n, updated)
        for pos in updated:
            self.surface._mark_mipmap_dirty(*pos)
        bbox = lib.surface.get_tiles_bbox(updated)
        self.surface.notify_observers(*bbox)
        return blanks_remaining or moves_remaining

    def _process_moves(self, n, updated):
        """Internal: process pending tile moves

        :param int n: as for process()
        :param set updated: Set of tile indices to be redrawn (in+out)
        :returns: Whether moves need to be processed
        :rtype: bool

        """
        if self.chunks_i > len(self.chunks):
            return False
        if n <= 0:
            n = len(self.chunks)  # process all remaining
        is_integral = len(self.slices_x) == 1 and len(self.slices_y) == 1
        for src_t in self.chunks[self.chunks_i:self.chunks_i + n]:
            src_tx, src_ty = src_t
            src_tile = self.snapshot.tiledict[src_t]
            for slice_x in self.slices_x:
                (src_x0, src_x1), (targ_tdx, targ_x0, targ_x1) = slice_x
                for slice_y in self.slices_y:
                    (src_y0, src_y1), (targ_tdy, targ_y0, targ_y1) = slice_y
                    targ_tx = src_tx + targ_tdx
                    targ_ty = src_ty + targ_tdy
                    targ_t = targ_tx, targ_ty
                    if is_integral:
                        # We're lucky. Perform a straight data copy.
                        self.surface.tiledict[targ_t] = src_tile.copy()
                        updated.add(targ_t)
                        self.written.add(targ_t)
                        continue
                    # Get a tile to write
                    targ_tile = None
                    if targ_t in self.written:
                        # Reuse a target tile made earlier in this
                        # update cycle
                        targ_tile = self.surface.tiledict.get(targ_t, None)
                    if targ_tile is None:
                        # Create and store a new blank target tile
                        # to avoid corruption
                        targ_tile = _Tile()
                        self.surface.tiledict[targ_t] = targ_tile
                        self.written.add(targ_t)
                    # Copy this source slice to the destination
                    targ_tile.rgba[targ_y0:targ_y1, targ_x0:targ_x1] \
                        = src_tile.rgba[src_y0:src_y1, src_x0:src_x1]
                    updated.add(targ_t)
            # The source tile has been fully processed at this point,
            # and can be removed from the output dict if it hasn't
            # also been written to.
            if src_t in self.surface.tiledict and src_t not in self.written:
                self.surface.tiledict.pop(src_t, None)
                updated.add(src_t)
        # Move on, and return whether we're complete
        self.chunks_i += n
        return self.chunks_i < len(self.chunks)

    def _process_blanks(self, n, updated):
        """Internal: process blanking-out queue

        :param int n: as for process()
        :param set updated: Set of tile indices to be redrawn (in+out)
        :returns: Whether the blanking queue is empty
        :rtype: bool

        """
        if n <= 0:
            n = len(self.blank_queue)
        while len(self.blank_queue) > 0 and n > 0:
            t = self.blank_queue.pop(0)
            if t not in self.written:
                self.surface.tiledict.pop(t, None)
                updated.add(t)
                n -= 1
        return len(self.blank_queue) > 0


def calc_translation_slices(dc):
    """Returns a list of offsets and slice extents for a translation

    :param dc: translation amount along the axis of interest (pixels)
    :type dc: int
    :returns: list of offsets and slice extents

    The returned slice list's members are of the form

        ((src_c0, src_c1), (targ_tdc, targ_c0, targ_c1))

    where ``src_c0`` and ``src_c1`` determine the extents of the source
    slice within a tile, their ``targ_`` equivalents specify where to
    put that slice in the target tile, and ``targ_tdc`` is the tile
    offset. For example,

        >>> assert N == 64, "FIXME: test only valid for 64 pixel tiles"
        >>> calc_translation_slices(N*2)
        [((0, 64), (2, 0, 64))]

    This indicates that all data from each tile is to be put exactly two
    tiles after the current tile index. In this case, a simple copy will
    suffice. Normally though, translations require slices.

        >>> calc_translation_slices(-16)
        [((0, 16), (-1, 48, 64)), ((16, 64), (0, 0, 48))]

    Two slices are needed for each tile: one strip of 16 pixels at the
    start to be copied to the end of output tile immediately before the
    current tile, and one strip of 48px to be copied to the start of the
    output tile having the same as the input.

    """
    dcr = dc % N
    tdc = (dc // N)
    if dcr == 0:
        return [
            ((0, N), (tdc, 0, N))
        ]
    else:
        return [
            ((0, N-dcr), (tdc, dcr, N)),
            ((N-dcr, N), (tdc+1, 0, dcr))
        ]


# Set which surface backend to use
Surface = MyPaintSurface


def _new_backend_surface():
    """Fetches a new backend surface object for C test code to use.

    Used by mypaintlib internals during tests: see lib/tiledsurface.hpp.
    The resultant pointer, after swizzling with SWIG_ConvertPtr(),
    exposes the libmypaint "MyPaintSurface" interface.

    """
    surface = Surface()
    return surface.backend


class BackgroundError(Exception):
    """Errors raised by Background during failed initiailizations"""
    pass


class Background (Surface):
    """A background layer surface, with a repeating image"""

    def __init__(self, obj, mipmap_level=0):
        """Construct from a color or from a NumPy array

        :param obj: RGB triple (uint8), or a HxWx4 or HxWx3 numpy array which
           can be either uint8 or uint16.
        :param mipmap_level: mipmap level, used internally. Root is zero.
        """

        if not isinstance(obj, np.ndarray):
            r, g, b = obj
            obj = np.zeros((N, N, 3), dtype='uint8')
            obj[:, :, :] = r, g, b

        height, width = obj.shape[0:2]
        if height % N or width % N:
            raise BackgroundError(
                "unsupported background tile size: %dx%d"
                % (width, height),
            )

        super(Background, self).__init__(mipmap_level=0, looped=True,
                                         looped_size=(width, height))
        self.load_from_numpy(obj, 0, 0)

        # Generate mipmap
        if mipmap_level <= MAX_MIPMAP_LEVEL:
            mipmap_obj = np.zeros((height, width, 4), dtype='uint16')
            for ty in range(height // N * 2):
                for tx in range(width // N * 2):
                    with self.tile_request(tx, ty, readonly=True) as src:
                        mypaintlib.tile_downscale_rgba16(src, mipmap_obj,
                                                         tx * N // 2,
                                                         ty * N // 2)

            self.mipmap = Background(mipmap_obj, mipmap_level+1)
            self.mipmap.parent = self
            self.mipmap_level = mipmap_level

    def _create_mipmap_surfaces(self):
        """Internal override: Background uses a different mipmap impl."""
        return None

    def load_from_numpy(self, arr, x, y):
        """Loads tile data from a numpy array

        This extends the base class's implementation with additional support
        for tile-aligned uint16 data.

        """
        h, w, channels = arr.shape
        if h <= 0 or w <= 0:
            return (x, y, w, h)
        if arr.dtype == 'uint16':
            assert w % N == 0 and h % N == 0
            assert x == 0 and y == 0
            for ty in range(h // N):
                for tx in range(w // N):
                    with self.tile_request(tx, ty, readonly=False) as dst:
                        dst[:, :, :] = arr[ty*N:(ty+1)*N, tx*N:(tx+1)*N, :]
            return (x, y, w, h)
        else:
            return super(Background, self).load_from_numpy(arr, x, y)


def flood_fill(src, fill_args, dst):
    """Fills connected areas of one surface into another

    :param src: Source surface-like object
    :type src: Anything supporting readonly tile_request()
    :param fill_args: fill arguments object
    :type fill_args: lib.floodfill.FloodFillArguments
    :param dst: Target surface
    :type dst: MyPaintSurface

    See also `lib.layer.Layer.flood_fill()`.

    """
    lib.floodfill._EMPTY_RGBA = transparent_tile.rgba
    return lib.floodfill.flood_fill(src, fill_args, dst)


class PNGFileUpdateTask (object):
    """Piecemeal callable: writes to or replaces a PNG file

    See lib.autosave.Autosaveable.

    >>> from tempfile import mkdtemp
    >>> from shutil import rmtree
    >>> import os.path
    >>> surf = MyPaintSurface._mock()
    >>> tmpdir = mkdtemp(suffix="_pngupdate")
    >>> tmpfile = os.path.join(tmpdir, "test.png")
    >>> try:
    ...     updater = PNGFileUpdateTask(surf, tmpfile, surf.get_bbox(), True)
    ...     while updater():
    ...         logger.debug("Wrote a tile strip")
    ...     assert os.path.isfile(tmpfile)
    ...     assert os.path.getsize(tmpfile) > 0
    ... finally:
    ...     rmtree(tmpdir)

    """

    def __init__(self, surface, filename, rect, alpha,
                 single_tile_pattern=False,
                 save_srgb_chunks=False,
                 **kwargs):
        super(PNGFileUpdateTask, self).__init__()
        self._final_filename = filename
        # Sizes. Save at least one tile to allow empty docs to be written
        if not rect:
            rect = surface.get_bbox()
        x, y, w, h = rect
        if w == 0 or h == 0:
            x, y, w, h = (0, 0, 1, 1)
            rect = (x, y, w, h)
        # Snapshot and recreate
        clone_surface = Surface(
            looped=surface.looped,
            looped_size=surface.looped_size,
        )
        clone_surface.load_snapshot(surface.save_snapshot())
        # Open a tempfile for writing
        tmp_filename = filename + ".tmp"
        if os.path.exists(tmp_filename):
            os.unlink(tmp_filename)
        tmp_fp = open(tmp_filename, "wb")
        self._png_writer = mypaintlib.ProgressivePNGWriter(
            tmp_fp,
            w, h,
            alpha,
            save_srgb_chunks,
        )
        self._tmp_filename = tmp_filename
        self._tmp_fp = tmp_fp
        # What to write
        self._strips_iter = lib.surface.scanline_strips_iter(
            clone_surface, rect, alpha=alpha,
            single_tile_pattern=single_tile_pattern,
            **kwargs
        )
        logger.debug("autosave: scheduled update of %r", self._final_filename)

    def __call__(self, *args, **kwargs):
        if not (self._png_writer and self._strips_iter):
            raise RuntimeError("Called too many times")
        try:
            strip = next(self._strips_iter)
            self._png_writer.write(strip)
            return True
        except StopIteration:
            try:
                self._png_writer.close()
            except Exception:
                logger.exception(
                    "Caught PNG writer close() exception "
                    "in StopIteration handler",
                )
            self._png_writer = None
            self._strips_iter = None
            self._tmp_fp.close()
            lib.fileutils.replace(
                self._tmp_filename,
                self._final_filename,
            )
            logger.debug("autosave: updated %r", self._final_filename)
            return False
        except Exception:
            try:
                self._png_writer.close()
            except Exception:
                logger.exception(
                    "Caught PNG writer close() exception "
                    "in general Exception handler handler",
                )
            self._png_writer = None
            self._strips_iter = None
            self._tmp_fp.close()
            if os.path.exists(self._tmp_filename):
                os.unlink(self._tmp_filename)
            logger.error("Original exception will be raised normally.")
            raise


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# This file is part of MyPaint.
# Copyright (C) 2019 by The MyPaint Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from lib.observable import event


class BlendMode(object):
    """A local on-off switch for a blend mode, with a change event"""

    def __init__(self, name, setting_name, active=False):
        self.name = name
        self.setting_name = setting_name
        self._active = active
        self.enabled = True

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, active):
        old_active = self._active
        self._active = active
        if old_active != active:
            self.changed()

    @event
    def changed(self):
        """Event dispatched triggered by a
        change in activation status"""


class BlendModes(object):
    """Proxy values for tools with individual blend mode states

    Used by tool modes to maintain their own instances of active
    blend modes and blend mode history.

    In order to enable individual blend modes for a tool mode,
    simply create its own BlendModes instance in a property named
    blend_modes and bind any listeners to it as needed.

    This class ensures that only a single mode is active at any one time
    and maintains a stack of modes to determine which it should switch to
    on deactivation.
    """

    NORMAL = 0
    ERASE = 1
    LOCK_ALPHA = 2
    COLORIZE = 3

    def __init__(self):
        self.eraser_mode = BlendMode("eraser_mode", "eraser")
        self.eraser_mode.mode_type = self.ERASE
        self.lock_alpha_mode = BlendMode("lock_alpha_mode", "lock_alpha")
        self.lock_alpha_mode.mode_type = self.LOCK_ALPHA
        self.colorize_mode = BlendMode("colorize_mode", "colorize")
        self.colorize_mode.mode_type = self.COLORIZE
        self.normal_mode = BlendMode("normal_mode", None, True)
        self.normal_mode.mode_type = self.NORMAL

        self.modes = [
            self.eraser_mode, self.lock_alpha_mode,
            self.colorize_mode, self.normal_mode
        ]
        self.mode_names = map(lambda x: x.name, self.modes)
        # Keep track of the active mode
        self.active_mode = self.normal_mode
        self.history = []

        for m in self.modes:
            m.changed += self._update

    def _push_history(self, mode):
        if mode is self.normal_mode:
            del self.history[:]
            return
        if mode in self.history:
            self.history.remove(mode)
        self.history.append(mode)

    def _pop_history(self, removed):
        while len(self.history) > 0:
            mode = self.history.pop()
            if mode is not removed:
                mode.active = True
                return
        self.normal_mode.active = True

    def _update(self, mode):
        old = self.active_mode
        old._active = False
        if mode.active:
            self._push_history(mode)
            self.active_mode = mode
            self.mode_changed(old, mode)
        else:
            self._pop_history(mode)

    @event
    def mode_changed(self, old_mode, new_mode):
        """Triggers when a mode changes,
        passing an instance of the changed mode"""


class BlendModeManager(object):
    """Manages blend mode actions by updating and redirecting
    callbacks based on BlendModes state models.

    A single instance "blendmodemanager" of this class is
    accessible from the Application singleton.

    This class allows multiple tool modes to make use of the same
    blend mode actions (same hotkeys etc.) by maintaining their
    own state in a BlendModes object, and taking control of the actions
    by registering that state when necessary.
    """

    def __init__(self, app):
        self.app = app
        self.delegates = []
        self.action_group = self.app.builder.get_object("BrushModifierActions")
        self.eraser_mode = app.find_action("BlendModeEraser")
        self.lock_alpha_mode = app.find_action("BlendModeLockAlpha")
        self.normal_mode = app.find_action("BlendModeNormal")
        self.colorize_mode = app.find_action("BlendModeColorize")
        self.actions = {
            "eraser_mode": self.eraser_mode,
            "lock_alpha_mode": self.lock_alpha_mode,
            "normal_mode": self.normal_mode,
            "colorize_mode": self.colorize_mode
        }

        for name in self.actions:
            action = self.actions[name]
            action.set_draw_as_radio(True)
            self.app.kbm.takeover_action(action)

        self._bm = None

    def register(self, bm):
        """Connect the blend mode actions to the given BlendModes object"""
        assert isinstance(bm, BlendModes)
        self.delegates.insert(0, bm)
        self._setup(bm)

    def update(self, bm, old, new):
        """Update actions without triggering change listeners"""
        old_action = self.actions[old.name]
        old_action.block_activate()
        old_action.set_active(False)
        old_action.unblock_activate()

        new_action = self.actions[new.name]
        new_action.block_activate()
        new_action.set_active(True)
        new_action.unblock_activate()

    def _setup(self, bm):
        """Set up listener and controls for new model and
        remove listener for old model"""
        # Deregister old change listener
        if self._bm:
            self._bm.mode_changed -= self.update
        # Update change listener
        self._bm = bm
        self._bm.mode_changed += self.update
        # Set up actions based on new BlendModes object
        for mode in bm.modes:
            action = self.actions[mode.name]
            action.block_activate()
            action.set_active(mode.active)
            action.set_sensitive(mode.enabled)
            action.unblock_activate()

    def deregister(self, bm):
        """Disconnect the blend mode actions from the given BlendModes
        if it is active.

        Remove the object from the stack and connect the next object in
        line for control, if such an object exists.
        """
        assert(isinstance(bm, BlendModes))
        if bm in self.delegates:
            self.delegates.remove(bm)
        else:
            return
        if self.delegates:
            self._setup(self.delegates[0])
        else:
            self._bm = None
            for action in self.actions.values():
                action.set_active(False)
                action.set_enabled(False)

    def blend_mode_normal_cb(self, action):
        if self._bm:
            self._bm.normal_mode.active = action.get_active()

    def blend_mode_eraser_cb(self, action):
        if self._bm:
            self._bm.eraser_mode.active = action.get_active()

    def blend_mode_lock_alpha_cb(self, action):
        if self._bm:
            self._bm.lock_alpha_mode.active = action.get_active()

    def blend_mode_colorize_cb(self, action):
        if self._bm:
            self._bm.colorize_mode.active = action.get_active()



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2018 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Helpers, and constants for the XML dialects MyPaint uses."""

## Imports

from __future__ import absolute_import, division, print_function

from lib.pycompat import PY3

import xml.etree.ElementTree as _ET

## Consts for XML dialects
# Namespaces are registered by importing this module.

OPENRASTER_MEDIA_TYPE = "image/openraster"
OPENRASTER_VERSION = u"0.0.5"
OPENRASTER_MYPAINT_NS = "http://mypaint.org/ns/openraster"

_OPENRASTER_NAMESPACES = {
    "mypaint": OPENRASTER_MYPAINT_NS,
}
for prefix, uri in _OPENRASTER_NAMESPACES.items():
    _ET.register_namespace(prefix, uri)


## Helper functions

def indent_etree(elem, level=0):
    """Indent an XML etree.

    This does not seem to come with python?
    Source: http://effbot.org/zone/element-lib.htm#prettyprint

    """
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent_etree(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def xsd2bool(arg):
    """Converts an XSD boolean datatype string from XML to a Python bool

      >>> xsd2bool("true")
      True
      >>> xsd2bool("false")
      False
      >>> xsd2bool(1)
      True
      >>> xsd2bool(0)
      False

    Ref: http://www.w3.org/TR/xmlschema-2/#boolean

    """
    return str(arg).lower() in ("true", "1")


def escape(u, quot=False, apos=False):
    """Escapes a Unicode string for use in XML/HTML.

      >>> u = u'<foo> & "bar"'
      >>> escape(u)
      '&lt;foo&gt; &amp; "bar"'
      >>> escape(u, quot=True)
      '&lt;foo&gt; &amp; &quot;bar&quot;'
      >>> escape(None) is None
      True

    Works like ``cgi.escape()``, but adds character ref encoding for
    characters which lie outside the ASCII range.
    The returned str is ASCII.

    """
    if u is None:
        return None
    u = u.replace("&", "&amp;")
    u = u.replace("<", "&lt;")
    u = u.replace(">", "&gt;")
    if apos:
        u = u.replace("'", "&apos;")
    if quot:
        u = u.replace('"', "&quot;")
    s = u.encode("ascii", "xmlcharrefreplace")
    if PY3:
        s = s.decode("ascii")
    return s


## Module testing


def _test(self):
    import doctest
    doctest.testmod()


if __name__ == "__main__":
    _test()



# This file is part of MyPaint.
# Copyright (C) 2013-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Widgets and popup dialogs for making quick choices"""

## Imports

from __future__ import division, print_function
import abc

from lib.gibindings import Gtk

from .pixbuflist import PixbufList
from . import brushmanager
from . import brushselectionwindow
from . import widgets
from . import spinbox
from . import windowing
from lib.observable import event
import gui.colortools
from lib.pycompat import add_metaclass


## Module consts

_DEFAULT_PREFS_ID = u"default"


## Interfaces

@add_metaclass(abc.ABCMeta)
class Advanceable:
    """Interface for choosers which can be advanced by pressing keys.

    Advancing happens if the chooser is already visible and its key is
    pressed again.  This can happen repeatedly.  The actual action
    performed is up to the implementation: advancing some some choosers
    may move them forward through pages of alternatives, while other
    choosers may actually change a brush setting as they advance.

    """

    @abc.abstractmethod
    def advance(self):
        """Advances the chooser to the next page or choice.

        Choosers should remain open when their advance() method is
        invoked. The actual action performed is up to the concrete
        implementation: see the class docs.

        """


## Class defs

class QuickBrushChooser (Gtk.VBox):
    """A quick chooser widget for brushes"""

    ## Class constants

    _PREFS_KEY_TEMPLATE = u"brush_chooser.%s.selected_group"
    ICON_SIZE = 48

    ## Method defs

    def __init__(self, app, prefs_id=_DEFAULT_PREFS_ID):
        """Initialize"""
        Gtk.VBox.__init__(self)
        self.app = app
        self.bm = app.brushmanager

        self._prefs_key = self._PREFS_KEY_TEMPLATE % (prefs_id,)
        active_group_name = app.preferences.get(self._prefs_key, None)

        model = self._make_groups_sb_model()
        self.groups_sb = spinbox.ItemSpinBox(model, self._groups_sb_changed_cb,
                                             active_group_name)
        active_group_name = self.groups_sb.get_value()

        brushes = self.bm.get_group_brushes(active_group_name)

        self.brushlist = PixbufList(
            brushes, self.ICON_SIZE, self.ICON_SIZE,
            namefunc=brushselectionwindow.managedbrush_namefunc,
            pixbuffunc=brushselectionwindow.managedbrush_pixbuffunc,
            idfunc=brushselectionwindow.managedbrush_idfunc
        )
        self.brushlist.dragging_allowed = False
        self.bm.groups_changed += self._groups_changed_cb
        self.bm.brushes_changed += self._brushes_changed_cb
        self.brushlist.item_selected += self._item_selected_cb

        scrolledwin = Gtk.ScrolledWindow()
        scrolledwin.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS)
        scrolledwin.add(self.brushlist)
        w = int(self.ICON_SIZE * 4.5)
        h = int(self.ICON_SIZE * 5.0)
        scrolledwin.set_min_content_width(w)
        scrolledwin.set_min_content_height(h)
        scrolledwin.get_child().set_size_request(w, h)

        self.pack_start(self.groups_sb, False, False, 0)
        self.pack_start(scrolledwin, True, True, 0)
        self.set_spacing(widgets.SPACING_TIGHT)

    def _item_selected_cb(self, pixbuf_list, brush):
        """Internal: call brush_selected event when an item is chosen"""
        self.brush_selected(brush)

    @event
    def brush_selected(self, brush):
        """Event: a brush was selected

        :param brush: The newly chosen brush
        """

    def _make_groups_sb_model(self):
        """Internal: create the model for the group choice spinbox"""
        group_names = sorted(self.bm.groups.keys())
        model = []
        for name in group_names:
            label_text = brushmanager.translate_group_name(name)
            model.append((name, label_text))
        return model

    def _groups_changed_cb(self, bm):
        """Internal: update the spinbox model at the top of the widget"""
        model = self._make_groups_sb_model()
        self.groups_sb.set_model(model)
        # In case the group has been deleted and recreated, we do this:
        group_name = self.groups_sb.get_value()
        group_brushes = self.bm.groups.get(group_name, [])
        self.brushlist.itemlist = group_brushes
        self.brushlist.update()
        # See https://github.com/mypaint/mypaint/issues/654

    def _brushes_changed_cb(self, bm, brushes):
        """Internal: update the PixbufList if its group was changed."""
        # CARE: this might be called in response to the group being deleted.
        # Don't recreate it by accident.
        group_name = self.groups_sb.get_value()
        group_brushes = self.bm.groups.get(group_name)
        if brushes is group_brushes:
            self.brushlist.update()

    def _groups_sb_changed_cb(self, group_name):
        """Internal: update the list of brush icons when the group changes"""
        self.app.preferences[self._prefs_key] = group_name
        group_brushes = self.bm.groups.get(group_name, [])
        self.brushlist.itemlist = group_brushes
        self.brushlist.update()

    def advance(self):
        """Advances to the next page of brushes."""
        self.groups_sb.next()


class BrushChooserPopup (windowing.ChooserPopup):
    """Speedy brush chooser popup"""

    def __init__(self, app, prefs_id=_DEFAULT_PREFS_ID):
        """Initialize.

        :param gui.application.Application app: main app instance
        :param unicode prefs_id: prefs identifier for the chooser

        The prefs identifier forms part of preferences key which store
        layout and which page of the chooser is selected. It should
        follow the same syntax rules as Python simple identifiers.

        """
        windowing.ChooserPopup.__init__(
            self,
            app = app,
            actions = [
                'ColorChooserPopup',
                'ColorChooserPopupFastSubset',
                'BrushChooserPopup',
            ],
            config_name = "brush_chooser.%s" % (prefs_id,),
        )
        self._chosen_brush = None
        self._chooser = QuickBrushChooser(app, prefs_id=prefs_id)
        self._chooser.brush_selected += self._brush_selected_cb

        bl = self._chooser.brushlist
        bl.connect("button-release-event", self._brushlist_button_release_cb)

        self.add(self._chooser)

    def _brush_selected_cb(self, chooser, brush):
        """Internal: update the response brush when an icon is clicked"""
        self._chosen_brush = brush

    def _brushlist_button_release_cb(self, *junk):
        """Internal: send an accept response on a button release

        We only send the response (and close the dialog) on button release to
        avoid accidental dabs with the stylus.
        """
        if self._chosen_brush is not None:
            bm = self.app.brushmanager
            bm.select_brush(self._chosen_brush)
            self.hide()
            self._chosen_brush = None

    def advance(self):
        """Advances to the next page of brushes."""
        self._chooser.advance()


class QuickColorChooser (Gtk.VBox):
    """A quick chooser widget for colors"""

    ## Class constants
    _PREFS_KEY_TEMPLATE = u"color_chooser.%s.selected_adjuster"
    _ALL_ADJUSTER_CLASSES = [
        gui.colortools.HCYWheelTool,
        gui.colortools.HSVWheelTool,
        gui.colortools.PaletteTool,
        gui.colortools.HSVCubeTool,
        gui.colortools.HSVSquareTool,
        gui.colortools.ComponentSlidersTool,
        gui.colortools.RingsColorChangerTool,
        gui.colortools.WashColorChangerTool,
        gui.colortools.CrossedBowlColorChangerTool,
    ]
    _SINGLE_CLICK_ADJUSTER_CLASSES = [
        gui.colortools.PaletteTool,
        gui.colortools.WashColorChangerTool,
        gui.colortools.CrossedBowlColorChangerTool,
    ]

    def __init__(self, app, prefs_id=_DEFAULT_PREFS_ID, single_click=False):
        Gtk.VBox.__init__(self)
        self._app = app
        self._spinbox_model = []
        self._adjs = {}
        self._pages = []
        mgr = app.brush_color_manager
        if single_click:
            adjuster_classes = self._SINGLE_CLICK_ADJUSTER_CLASSES
        else:
            adjuster_classes = self._ALL_ADJUSTER_CLASSES
        for page_class in adjuster_classes:
            name = page_class.__name__
            page = page_class()
            self._pages.append(page)
            self._spinbox_model.append((name, page.tool_widget_title))
            self._adjs[name] = page
            page.set_color_manager(mgr)
            if page_class in self._SINGLE_CLICK_ADJUSTER_CLASSES:
                page.connect_after(
                    "button-release-event",
                    self._ccwidget_btn_release_cb,
                )
        self._prefs_key = self._PREFS_KEY_TEMPLATE % (prefs_id,)
        active_page = app.preferences.get(self._prefs_key, None)
        sb = spinbox.ItemSpinBox(self._spinbox_model, self._spinbox_changed_cb,
                                 active_page)
        active_page = sb.get_value()
        self._spinbox = sb
        self._active_adj = self._adjs[active_page]
        self.pack_start(sb, False, False, 0)
        self.pack_start(self._active_adj, True, True, 0)
        self.set_spacing(widgets.SPACING_TIGHT)

    def _spinbox_changed_cb(self, page_name):
        self._app.preferences[self._prefs_key] = page_name
        self.remove(self._active_adj)
        new_adj = self._adjs[page_name]
        self._active_adj = new_adj
        self.pack_start(self._active_adj, True, True, 0)
        self._active_adj.show_all()

    def _ccwidget_btn_release_cb(self, ccwidget, event):
        """Internal: fire "choice_completed" after clicking certain widgets"""
        self.choice_completed()
        return False

    @event
    def choice_completed(self):
        """Event: a complete selection was made

        This is emitted by button-release events on certain kinds of colour
        chooser page. Not every page in the chooser emits this event, because
        colour is a three-dimensional quantity: clicking on a two-dimensional
        popup can't make a complete choice of colour with most pages.

        The palette page does emit this event, and it's the default.
        """

    def advance(self):
        """Advances to the next color selector."""
        self._spinbox.next()


class ColorChooserPopup (windowing.ChooserPopup):
    """Speedy color chooser dialog"""

    def __init__(self, app, prefs_id=_DEFAULT_PREFS_ID, single_click=False):
        """Initialize.

        :param gui.application.Application app: main app instance
        :param unicode prefs_id: prefs identifier for the chooser
        :param bool single_click: limit to just the single-click adjusters

        The prefs identifier forms part of preferences key which store
        layout and which page of the chooser is selected. It should
        follow the same syntax rules as Python simple identifiers.

        """
        windowing.ChooserPopup.__init__(
            self,
            app = app,
            actions = [
                'ColorChooserPopup',
                'ColorChooserPopupFastSubset',
                'BrushChooserPopup',
            ],
            config_name = u"color_chooser.%s" % (prefs_id,),
        )
        self._chooser = QuickColorChooser(
            app,
            prefs_id=prefs_id,
            single_click=single_click,
        )
        self._chooser.choice_completed += self._choice_completed_cb
        self.add(self._chooser)

    def _choice_completed_cb(self, chooser):
        """Internal: close when a choice is (fully) made

        Close the dialog on button release only to avoid accidental dabs
        with the stylus.
        """
        self.hide()

    def advance(self):
        """Advances to the next color selector."""
        self._chooser.advance()


## Classes: interface registration

Advanceable.register(QuickBrushChooser)
Advanceable.register(QuickColorChooser)
Advanceable.register(BrushChooserPopup)
Advanceable.register(ColorChooserPopup)



# This file is part of MyPaint.
# Copyright (C) 2012-2013 by Andrew Chadwick <a.t.chadwickgmail.com>
# Copyright (C) 2014-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Axis-aligned planar slice of an HSV color cube, and a depth slider.
"""

from __future__ import division, print_function

from gettext import gettext as _

import cairo
from lib.gibindings import Gtk

from .util import clamp
from .util import draw_marker_circle
from lib.color import HSVColor
from lib.color import RGBColor
from .adjbases import ColorAdjusterWidget
from .adjbases import ColorAdjuster
from .adjbases import SliderColorAdjuster
from .adjbases import IconRenderableColorAdjusterWidget
from .combined import CombinedAdjusterPage
from .uimisc import borderless_button
from .uimisc import PRIMARY_ADJUSTERS_MIN_WIDTH
from .uimisc import PRIMARY_ADJUSTERS_MIN_HEIGHT

from lib.pycompat import xrange


class HSVCubePage (CombinedAdjusterPage):
    """Slice+depth view through an HSV cube: page for `CombinedAdjuster`.

    The page includes a button for tumbling the cube, i.e. changing which of
    the color components the slice and the depth slider refer to.

    """

    # Tooltip mappings, indexed by whatever the slider currently represents
    _slider_tooltip_map = dict(h=_("HSV Hue"),
                               s=_("HSV Saturation"),
                               v=_("HSV Value"))
    _slice_tooltip_map = dict(h=_("HSV Saturation and Value"),
                              s=_("HSV Hue and Value"),
                              v=_("HSV Hue and Saturation"))

    def __init__(self):
        self._faces = ['h', 's', 'v']
        button = borderless_button(
            icon_name='mypaint-hsv-rotate-symbolic',
            size=Gtk.IconSize.MENU,
            tooltip=_("Rotate cube (show different axes)")
        )
        button.connect("clicked", lambda *a: self.tumble())
        self.__slice = HSVCubeSlice(self)
        self.__slider = HSVCubeSlider(self)
        s_align = Gtk.Alignment(xalign=0.5, yalign=0, xscale=0, yscale=1)
        s_align.add(self.__slider)

        table = Gtk.Table(n_rows=2, n_columns=2)

        xopts = Gtk.AttachOptions.FILL | Gtk.AttachOptions.EXPAND
        yopts = Gtk.AttachOptions.FILL | Gtk.AttachOptions.EXPAND

        table.attach(s_align, 0, 1, 0, 1, Gtk.AttachOptions.FILL, yopts, 3, 3)
        table.attach(
            button, 0, 1, 1, 2,
            Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 3, 3)
        table.attach(self.__slice, 1, 2, 0, 2, xopts, yopts, 3, 3)
        self.__table = table
        self._update_tooltips()

    @classmethod
    def get_page_icon_name(self):
        return 'mypaint-tool-hsvcube'

    @classmethod
    def get_page_title(self):
        return _('HSV Cube')

    @classmethod
    def get_page_description(self):
        return _("An HSV cube which can be rotated to show different "
                 "planar slices.")

    def get_page_widget(self):
        return self.__table

    def tumble(self):
        f0 = self._faces.pop(0)
        self._faces.append(f0)
        self.__slider.queue_draw()
        self.__slice.queue_draw()
        self._update_tooltips()

    def _update_tooltips(self):
        f0 = self._faces[0]
        self.__slice.set_tooltip_text(self._slice_tooltip_map[f0])
        self.__slider.set_tooltip_text(self._slider_tooltip_map[f0])

    def set_color_manager(self, manager):
        ColorAdjuster.set_color_manager(self, manager)
        self.__slider.set_color_manager(manager)
        self.__slice.set_color_manager(manager)


class HSVCubeSlider (SliderColorAdjuster):
    """Depth of the planar slice of a cube.
    """

    vertical = True
    samples = 4

    def __init__(self, cube):
        SliderColorAdjuster.__init__(self)
        self.__cube = cube

    def get_background_validity(self):
        col = HSVColor(color=self.get_managed_color())
        f0, f1, f2 = self.__cube._faces
        return f0, getattr(col, f1), getattr(col, f2)

    def get_color_for_bar_amount(self, amt):
        col = HSVColor(color=self.get_managed_color())
        f0 = self.__cube._faces[0]
        setattr(col, f0, amt)
        return col

    def get_bar_amount_for_color(self, col):
        f0 = self.__cube._faces[0]
        amt = getattr(col, f0)
        return amt


class HSVCubeSlice (IconRenderableColorAdjusterWidget):
    """Planar slice through an HSV cube.
    """

    def __init__(self, cube):
        ColorAdjusterWidget.__init__(self)
        w = PRIMARY_ADJUSTERS_MIN_WIDTH
        h = PRIMARY_ADJUSTERS_MIN_HEIGHT
        self.set_size_request(w, h)
        self.__cube = cube

    def __get_faces(self):
        f1 = self.__cube._faces[1]
        f2 = self.__cube._faces[2]
        if f2 == 'h':
            f1, f2 = f2, f1
        return f1, f2

    def render_background_cb(self, cr, wd, ht, icon_border=None):
        col = HSVColor(color=self.get_managed_color())
        b = icon_border
        if b is None:
            b = self.BORDER_WIDTH
        eff_wd = int(wd - 2*b)
        eff_ht = int(ht - 2*b)
        f1, f2 = self.__get_faces()

        step = max(1, int(eff_wd // 128))

        rect_x, rect_y = int(b)+0.5, int(b)+0.5
        rect_w, rect_h = int(eff_wd)-1, int(eff_ht)-1

        # Paint the central area offscreen
        cr.push_group()
        for x in xrange(0, eff_wd, step):
            amt = x / eff_wd
            setattr(col, f1, amt)
            setattr(col, f2, 1.0)
            lg = cairo.LinearGradient(b+x, b, b+x, b+eff_ht)
            lg.add_color_stop_rgb(*([0.0] + list(col.get_rgb())))
            setattr(col, f2, 0.0)
            lg.add_color_stop_rgb(*([1.0] + list(col.get_rgb())))
            cr.rectangle(b+x, b, step, eff_ht)
            cr.set_source(lg)
            cr.fill()
        slice_patt = cr.pop_group()

        # Tango-like outline
        cr.set_line_join(cairo.LINE_JOIN_ROUND)
        cr.rectangle(rect_x, rect_y, rect_w, rect_h)
        cr.set_line_width(self.OUTLINE_WIDTH)
        cr.set_source_rgba(*self.OUTLINE_RGBA)
        cr.stroke()

        # The main area
        cr.set_source(slice_patt)
        cr.paint()

        # Tango-like highlight over the top
        cr.rectangle(rect_x, rect_y, rect_w, rect_h)
        cr.set_line_width(self.EDGE_HIGHLIGHT_WIDTH)
        cr.set_source_rgba(*self.EDGE_HIGHLIGHT_RGBA)
        cr.stroke()

    def get_background_validity(self):
        col = HSVColor(color=self.get_managed_color())
        f0 = self.__cube._faces[0]
        return f0, getattr(col, f0)

    def get_color_at_position(self, x, y):
        alloc = self.get_allocation()
        b = self.BORDER_WIDTH
        wd = alloc.width
        ht = alloc.height
        eff_wd = wd - 2*b
        eff_ht = ht - 2*b
        f1_amt = clamp((x-b) / eff_wd, 0, 1)
        f2_amt = clamp((y-b) / eff_ht, 0, 1)
        col = HSVColor(color=self.get_managed_color())
        f1, f2 = self.__get_faces()
        f2_amt = 1.0 - f2_amt
        setattr(col, f1, f1_amt)
        setattr(col, f2, f2_amt)
        return col

    def get_position_for_color(self, col):
        col = HSVColor(color=col)
        f1, f2 = self.__get_faces()
        f1_amt = getattr(col, f1)
        f2_amt = getattr(col, f2)
        f2_amt = 1.0 - f2_amt
        alloc = self.get_allocation()
        b = self.BORDER_WIDTH
        wd = alloc.width
        ht = alloc.height
        eff_wd = wd - 2*b
        eff_ht = ht - 2*b
        x = b + f1_amt*eff_wd
        y = b + f2_amt*eff_ht
        return x, y

    def paint_foreground_cb(self, cr, wd, ht):
        x, y = self.get_position_for_color(self.get_managed_color())
        draw_marker_circle(cr, x, y)


if __name__ == '__main__':
    import os
    import sys
    from adjbases import ColorManager
    mgr = ColorManager(prefs={}, datapath='.')
    cube = HSVCubePage()
    cube.set_color_manager(mgr)
    mgr.set_color(RGBColor(0.3, 0.6, 0.7))
    if len(sys.argv) > 1:
        slice = HSVCubeSlice(cube)
        slice.set_color_manager(mgr)
        icon_name = cube.get_page_icon_name()
        for dir_name in sys.argv[1:]:
            slice.save_icon_tree(dir_name, icon_name)
    else:
        # Interactive test
        window = Gtk.Window()
        window.add(cube.get_page_widget())
        window.set_title(os.path.basename(sys.argv[0]))
        window.connect("destroy", lambda *a: Gtk.main_quit())
        window.show_all()
        Gtk.main()



# This file is part of MyPaint.
# Copyright (C) 2012-2014 by Andrew Chadwick <andrewc-git@piffle.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Base widgets for the colour selector module"""

from __future__ import division, print_function

import logging

from lib.gibindings import Gtk
import cairo


logger = logging.getLogger(__name__)


class CachedBgWidgetMixin (object):
    """Provides widgets with a cached background and drawing callbacks

    The background is invalidated whenever the size changes,
    or when the overridable background-validity method
    returns something new.
    """

    def __init__(self):
        """Instantiate, binding events

        This must only be called *after* the base class initializer.
        """
        self.connect("draw", self.__draw_cb)
        self.connect("size-allocate", self.__size_allocate_cb)
        self.__bg = None
        self.__bg_validity = None

    def __draw_cb(self, widget, cr):
        bg_valid = self.__bg is not None
        if bg_valid:
            validity = self.get_background_validity()
            if validity != self.__bg_validity:
                bg_valid = False
        if not bg_valid:
            self.__rerender_background()
            assert self.__bg is not None
        alloc = self.get_allocation()
        cr.set_source_surface(self.__bg, 0, 0)
        cr.paint()
        self.paint_foreground_cb(cr, alloc.width, alloc.height)

    def __size_allocate_cb(self, widget, alloc):
        self.clear_background()
        self.__rerender_background()

    def _get_background_size(self):
        alloc = self.get_allocation()
        return alloc.width, alloc.height

    def __rerender_background(self):
        w, h = self._get_background_size()
        surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
        cr = cairo.Context(surf)
        self.render_background_cb(cr, w, h)
        self.__bg = surf
        self.__bg_validity = self.get_background_validity()

    def get_background_validity(self):
        """Get a validity token for the cached background

        :return: a value which, if changed, means the bg must be redrawn

        Validity tokens may be anything comparable by ``==``, and
        should be simpler to compute than re-rendering the background.
        """
        raise NotImplementedError

    def render_background_cb(self, cr, w, h, *kw):
        """Render the background when needed"""
        raise NotImplementedError

    def paint_foreground_cb(self, cr, w, h):
        """Paints the foreground over the background """
        raise NotImplementedError

    def clear_background(self):
        """Clears the cached background, making it invalid"""
        self.__bg = None
        self.__bg_validity = None
        self.queue_draw()


class CachedBgDrawingArea (CachedBgWidgetMixin, Gtk.EventBox):
    """Base class for widgets with cached backgrounds"""

    def __init__(self):
        Gtk.EventBox.__init__(self)
        self.set_app_paintable(True)
        CachedBgWidgetMixin.__init__(self)


class IconRenderable (object):
    """Mixin for objects that can be rendered as a XDG icons

    Typically a cached icon file from disk will be quicker
    and more convenient than rendering the icon
    each time a tab needs to be drawn, so this mixin
    can be used for saving icons to disk too.
    """

    def render_as_icon(self, cr, size):
        """Renders as an icon into a Cairo context (unimplemented)

        The icon pixel size, `size`, is one of 48, 32, 24, 22 or 16.
        """
        raise NotImplementedError

    def save_icon_tree(self, dir_name, icon_name):
        """Saves a full set of XDG icons into a given root directory"""
        import os
        dpi = 90.0
        for size in (48, 32, 24, 22, 16, 'scalable'):
            if size == 'scalable':
                path = "%s/hicolor/scalable/actions" % (dir_name,)
                if not os.path.exists(path):
                    os.makedirs(path)
                filename = "%s/%s.svg" % (path, icon_name)
                s = 48.0
                scale = 72/dpi
                pts = s*scale
                surf = cairo.SVGSurface(filename, pts, pts)
                cr = cairo.Context(surf)
                cr.scale(scale, scale)
            else:
                path = "%s/hicolor/%dx%d/actions" % (dir_name, size, size)
                if not os.path.exists(path):
                    os.makedirs(path)
                filename = "%s/%s.png" % (path, icon_name)
                surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, size, size)
                s = size
                cr = cairo.Context(surf)
            self.render_as_icon(cr, size=int(s))
            surf.flush()
            if size != 'scalable':
                surf.write_to_png(filename)
            logger.info("rendered %r (size=%s)...", filename, size)



# This file is part of MyPaint.
# Copyright (C) 2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Function equivalents of (GLib) gettext's C macros.

Recommended usage:

  >>> from lib.gettext import C_
  >>> from lib.gettext import ngettext

Also supported, but mildly deprecated (consider C_ instead!):

  >>> from lib.gettext import gettext as _

Lots of older code uses ``from gettext import gettext as _``.
Don't do that in new code: pull in this module as ``lib.gettext``.
Importing this module should still work from within lib/ if code
still uses a relative import, however.

"""

from __future__ import absolute_import, division, print_function

from warnings import warn
from lib.gibindings import GLib

# Set the default encoding like PyGTK
from lib.pycompat import PY3
import sys
if not PY3:
    reload(sys)  # noqa: F821
    sys.setdefaultencoding("utf-8")

# Older code in lib imports these as "from gettext import gettext as _".
# Pull them in for backwards compat.
# Might change these to _Glib.dgettext/ngettext instead.

from gettext import gettext  # noqa: F401 E402
from gettext import ngettext  # noqa: F401 E402


# Newer code should use C_() even for simple cases, and provide contexts
# for translators.

def C_(context, msgid):  # noqa: N802
    """Mark a string for translation, with supplied context.

    :param str context: Disambiguating context.
    :param str msgid: String to translate.
    :returns: the translated Unicode string
    :rtype: str

    Convenience wrapper around g_dpgettext2. It's a function not a
    macro, but use it as if it was a C macro only: in other words, only
    use string literals so that the strings marked for translation can
    be extracted.

    Writing the context as a regular string literal and the string
    marked for translation as an explicit unicode lteral makes the fake
    macro easier to read in the code.

    """
    g_dpgettext2 = GLib.dpgettext2
    try:
        result = g_dpgettext2(None, context, msgid)
    except TypeError as e:
        # Expect "Argument 0 does not allow None as a value" sometimes.
        # This is a known problem with Ubuntu Server 12.04 when testing
        # lib - that version of g_dpgettext2() does not appear to allow
        # NULL for its first arg.
        wtmpl = "C_(): g_dpgettext2() raised %r. Try a newer GLib?"
        warn(
            wtmpl % (e,),
            RuntimeWarning,
            stacklevel = 1,
        )
        return msgid
    else:
        if isinstance(result, bytes):
            # GLib.dbgettext in PY2 does this, but PY3 doesn't. Weird.
            # Let's assume we can trust the Unicode string we get handed
            # in Python 3, and hope that this won't create a pile of
            # mojibake on screen.
            result = result.decode("utf-8")
        return result



# This file is part of MyPaint.
# Copyright (C) 2014-2016 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or


"""Modes for moving layers around on the canvas"""

## Imports
from __future__ import division, print_function

from gettext import gettext as _

from lib.gibindings import Gdk
from lib.gibindings import GLib

import gui.mode
import lib.command
import gui.cursor


## Class defs


class LayerMoveMode (gui.mode.ScrollableModeMixin,
                     gui.mode.DragMode):
    """Moving a layer interactively

    MyPaint is tile-based, and tiles must align between layers.
    Therefore moving layers involves copying data around. This is slow
    for very large layers, so the work is broken into chunks and
    processed in the idle phase of the GUI for greater responsiveness.

    """

    ## API properties and informational methods

    ACTION_NAME = 'LayerMoveMode'

    pointer_behavior = gui.mode.Behavior.CHANGE_VIEW
    scroll_behavior = gui.mode.Behavior.CHANGE_VIEW

    @classmethod
    def get_name(cls):
        return _(u"Move Layer")

    def get_usage(self):
        return _(u"Move the current layer")

    @property
    def active_cursor(self):
        cursor_name = gui.cursor.Name.HAND_CLOSED
        if not self._move_possible:
            cursor_name = gui.cursor.Name.FORBIDDEN_EVERYWHERE
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME,
            cursor_name,
        )

    @property
    def inactive_cursor(self):
        cursor_name = gui.cursor.Name.HAND_OPEN
        if not self._move_possible:
            cursor_name = gui.cursor.Name.FORBIDDEN_EVERYWHERE
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME,
            cursor_name,
        )

    permitted_switch_actions = set([
        'RotateViewMode',
        'ZoomViewMode',
        'PanViewMode',
    ] + gui.mode.BUTTON_BINDING_ACTIONS)

    ## Initialization

    def __init__(self, **kwds):
        super(LayerMoveMode, self).__init__(**kwds)
        self._cmd = None
        self._drag_update_idler_srcid = None
        self.final_modifiers = 0
        self._move_possible = False
        self._drag_active_tdw = None
        self._drag_active_model = None

    ## Layer stacking API

    def enter(self, doc, **kwds):
        super(LayerMoveMode, self).enter(doc, **kwds)
        self.final_modifiers = self.initial_modifiers
        rootstack = self.doc.model.layer_stack
        rootstack.current_path_updated += self._update_ui
        rootstack.layer_properties_changed += self._update_ui
        self._update_ui()

    def leave(self, **kwds):
        if self._cmd is not None:
            while self._finalize_move_idler():
                pass
        rootstack = self.doc.model.layer_stack
        rootstack.current_path_updated -= self._update_ui
        rootstack.layer_properties_changed -= self._update_ui
        return super(LayerMoveMode, self).leave(**kwds)

    def checkpoint(self, **kwds):
        """Commits any pending work to the command stack"""
        if self._cmd is not None:
            while self._finalize_move_idler():
                pass
        return super(LayerMoveMode, self).checkpoint(**kwds)

    ## Drag-mode API

    def drag_start_cb(self, tdw, event):
        """Drag initialization"""
        if self._move_possible and self._cmd is None:
            model = tdw.doc
            layer_path = model.layer_stack.current_path
            x0, y0 = tdw.display_to_model(self.start_x, self.start_y)
            cmd = lib.command.MoveLayer(model, layer_path, x0, y0)
            self._cmd = cmd
            self._drag_active_tdw = tdw
            self._drag_active_model = model
        return super(LayerMoveMode, self).drag_start_cb(tdw, event)

    def drag_update_cb(self, tdw, event, dx, dy):
        """UI and model updates during a drag"""
        if self._cmd:
            assert tdw is self._drag_active_tdw
            x, y = tdw.display_to_model(event.x, event.y)
            self._cmd.move_to(x, y)
            if self._drag_update_idler_srcid is None:
                idler = self._drag_update_idler
                self._drag_update_idler_srcid = GLib.idle_add(idler)
        return super(LayerMoveMode, self).drag_update_cb(tdw, event, dx, dy)

    def _drag_update_idler(self):
        """Processes tile moves in chunks as a background idler"""
        # Might have exited, in which case leave() will have cleaned up
        if self._cmd is None:
            self._drag_update_idler_srcid = None
            return False
        # Terminate if asked. Assume the asker will clean up.
        if self._drag_update_idler_srcid is None:
            return False
        # Process some tile moves, and carry on if there's more to do
        if self._cmd.process_move():
            return True
        self._drag_update_idler_srcid = None
        return False

    def drag_stop_cb(self, tdw):
        """UI and model updates at the end of a drag"""
        # Stop the update idler running on its next scheduling
        self._drag_update_idler_srcid = None
        # This will leave a non-cleaned-up move if one is still active,
        # so finalize it in its own idle routine.
        if self._cmd is not None:
            assert tdw is self._drag_active_tdw
            # Arrange for the background work to be done, and look busy
            tdw.set_sensitive(False)

            window = tdw.get_window()
            cursor = Gdk.Cursor.new_for_display(
                window.get_display(), Gdk.CursorType.WATCH)
            tdw.set_override_cursor(cursor)

            self.final_modifiers = self.current_modifiers()
            GLib.idle_add(self._finalize_move_idler)
        else:
            # Still need cleanup for tracking state, cursors etc.
            self._drag_cleanup()
        return super(LayerMoveMode, self).drag_stop_cb(tdw)

    def _finalize_move_idler(self):
        """Finalizes everything in chunks once the drag's finished"""
        if self._cmd is None:
            return False  # something else cleaned up
        while self._cmd.process_move():
            return True
        model = self._drag_active_model
        cmd = self._cmd
        tdw = self._drag_active_tdw
        self._cmd = None
        self._drag_active_tdw = None
        self._drag_active_model = None
        self._update_ui()
        tdw.set_sensitive(True)
        model.do(cmd)
        self._drag_cleanup()
        return False

    ## Helpers

    def _update_ui(self, *_ignored):
        """Updates the cursor, and the internal move-possible flag"""
        layer = self.doc.model.layer_stack.current
        self._move_possible = (
            layer.visible
            and not layer.locked
            and layer.branch_visible
            and not layer.branch_locked
        )
        self.doc.tdw.set_override_cursor(self.inactive_cursor)

    def _drag_cleanup(self):
        """Final cleanup after any drag is complete"""
        if self._drag_active_tdw:
            self._update_ui()  # update may have been deferred
        self._drag_active_tdw = None
        self._drag_active_model = None
        self._cmd = None
        if not self.doc:
            return
        if self is self.doc.modes.top:
            if self.initial_modifiers:
                if (self.final_modifiers & self.initial_modifiers) == 0:
                    self.doc.modes.pop()



# -*- coding: utf-8 -*-
# Copyright 2006 Joe Wreschnig
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation


"""I18n setup routines.

Stolen from QuodLibet (thanks, lazka!) <https://github.com/quodlibet
/quodlibet/blob/76bb75fa677a8a1db830fbe47d0000788184e903/quodlibet
/quodlibet/util/i18n.py#L58>.

"""

from __future__ import division, print_function

import os
import sys
import locale
import logging
logger = logging.getLogger(__name__)

USER_LOCALE_PREF = "lib.i18n.userlocale"


def getdefaultlocale():
    """Return default locale or None
    """
    try:
        return locale.getdefaultlocale()
    except Exception:
        logger.exception("Exception when getting default locale!")
        return None


def bcp47_to_language(code):
    """Takes a BCP 47 language identifier and returns a value suitable for the
    LANGUAGE env var.
    Only supports a small set of inputs and might return garbage..
    """

    if code == "zh-Hans":
        return "zh_CN"
    elif code == "zh-Hant":
        return "zh_TW"

    parts = code.split("-")

    # we only support ISO 639-1
    if not _is_iso_639_1(parts[0]):
        return parts[0].replace(":", "")
    lang_subtag = parts[0]

    region = ""
    if len(parts) >= 2 and _is_iso_639_1(parts[1]):
        region = parts[1]
    elif len(parts) >= 3 and _is_iso_639_1(parts[2]):
        region = parts[2]

    if region:
        return "%s_%s" % (lang_subtag, region)
    return lang_subtag


def _is_iso_639_1(s):
    return len(s) == 2 and s.isalpha()


def osx_locale_id_to_lang(id_):
    """Converts a NSLocale identifier to something suitable for LANG"""

    if "_" not in id_:
        return id_
    # id_ can be "zh-Hans_TW"
    parts = id_.rsplit("_", 1)
    ll = parts[0]
    ll = bcp47_to_language(ll).split("_")[0]
    return "%s_%s" % (ll, parts[1])


def set_i18n_envvars():
    """Set the LANG/LANGUAGE environment variables if not set.

    Covers the cases where the current platform doesn't use them by
    default (OS X, Windows). In these systems, it will pick up defaults
    by other OS-specific means.

    """

    if os.name == "nt":
        logger.debug(
            "Windows: figuring out fallbacks for gettext's "
            "LANG and LANGUAGE vars. Setting those will override "
            "this logic."
        )
        langs = []

        # Let the user's chosen UI language be the default, unless
        # overridden by a POSIX-style LANG or LANGUAGE var.
        try:
            import ctypes
            k32 = ctypes.windll.kernel32
            for lang_k32 in [k32.GetUserDefaultUILanguage(),
                             k32.GetSystemDefaultUILanguage()]:
                lang = locale.windows_locale.get(lang_k32)
                if lang not in langs:
                    langs.append(lang)
                    logger.debug("Windows: found UI language %r", lang)
        except:
            logger.exception("Windows: couldn't get UI language via ctypes")

        # On Win7 Professional, or Win7 Enterprise/Ultimate when there
        # are no UI language packs installed, this will pick up the
        # user's Region and Language → Formats setting when the code
        # above doesn't set anything up. Note that some users have to
        # fudge their "Formats" setting in order to get 3rd party
        # software to work without crashing.
        deflocale = getdefaultlocale()

        if deflocale:
            lang = deflocale[0]
            if lang not in langs:
                langs.append(lang)
                logger.debug(
                    "Windows: found fallback language %r, "
                    "inferred from Region & Languages -> Formats.",
                    lang,
                )

        # The POSIX-style LANG and LANGUAGE environment variables still
        # override all of the above. Since MSYS2 shells are going to use
        # them, that's reasonable.
        if langs:
            os.environ.setdefault('LANG', langs[0])
            os.environ.setdefault('LANGUAGE', ":".join(langs))
        logger.info("Windows: LANG=%r", os.environ.get("LANG"))
        logger.info("Windows: LANGUAGE=%r", os.environ.get("LANGUAGE"))

    elif sys.platform == "darwin":
        try:
            from AppKit import NSLocale
        except ImportError:
            logger.exception("OSX: failed to import AppKit.NSLocale")
            logger.warning("OSX: falling back to POSIX mechanisms.")
        else:
            logger.info(
                "OSX: imported AppKit.NSLocale OK. "
                "Will use for LANG, and for LANGUAGE order."
            )
            locale_id = NSLocale.currentLocale().localeIdentifier()
            lang = osx_locale_id_to_lang(locale_id)
            os.environ.setdefault('LANG', lang)
            preferred_langs = NSLocale.preferredLanguages()
            if preferred_langs:
                languages = map(bcp47_to_language, preferred_langs)
                os.environ.setdefault('LANGUAGE', ":".join(languages))
        logger.info("OSX: LANG=%r", os.environ.get("LANG"))
        logger.info("OSX: LANGUAGE=%r", os.environ.get("LANGUAGE"))

    else:
        logger.info("POSIX: LANG=%r", os.environ.get("LANG"))
        logger.info("POSIX: LANGUAGE=%r", os.environ.get("LANGUAGE"))


def fixup_i18n_envvars():
    """Sanitizes env vars before gettext can use them.

    LANGUAGE should support a priority list of languages with fallbacks,
    but doesn't work due to "en" not being known to gettext (This could
    be solved by providing a en.po in [QuodLibet] but all other
    libraries don't define it either). This tries to fix that.

    """

    try:
        langs = list(os.environ["LANGUAGE"].split(":"))
    except KeyError:
        return

    # So, this seems to be an undocumented feature where C selects "no
    # translation". Insert it into the priority list in an appropriate
    # place so gettext falls back to it there.

    if "en_US" in langs:
        # Our source language is US English… as written by non-US folks.
        # The best location for the "C" hack is right after that.
        i = langs.index("en_US")
        langs.insert(i + 1, "C")
    else:
        # Otherwise, insert it after the last non-US English lect we
        # find, so that "C" will override any subsequent other languages
        # for users who are happier with English.
        sanitized = []
        i = -1
        for lang in langs:
            sanitized.append(lang)
            if lang.startswith("en"):
                i = len(sanitized)
        if i >= 0:
            sanitized.insert(i, "C")
        langs = sanitized

    os.environ["LANGUAGE"] = ":".join(langs)
    logger.info("Value of LANGUAGE after cleanup: %r", os.environ["LANGUAGE"])



# This file is part of MyPaint.
# Copyright (C) 2019 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Functions and constants common to fill and morphological operations"""

import lib.helpers
import lib.mypaintlib
import numpy


N = lib.mypaintlib.TILE_SIZE

# Constant alpha value for fully opaque pixels
_OPAQUE = 1 << 15

# Keeping track of fully opaque tiles allows for potential
# substantial performance benefits for both morphological
# operations as well as feathering and compositing
_FULL_TILE = lib.mypaintlib.ConstTiles.ALPHA_OPAQUE()
_FULL_TILE.flags.writeable = False

# Keeping track of empty tiles (which are less likely to
# be produced during the fill) permits skipping the compositing
# step for these tiles
_EMPTY_TILE = lib.mypaintlib.ConstTiles.ALPHA_TRANSPARENT()
_EMPTY_TILE.flags.writeable = False


def new_full_tile(value, dimensions=(N, N), value_type='uint16'):
    """Return a new tile filled with the given value"""
    tile = numpy.empty(dimensions, value_type)
    tile.fill(value)
    return tile


def nine_grid(tile_coord):
    """ Return the input coordinate along with its neighbours.

    Return tile coordinates of the full nine-grid,
    relative to the input coordinate, in the following order:

    8 1 5
    4 0 2
    7 3 6
    """
    tile_x, tile_y = tile_coord
    offsets = [
        (0, 0), (0, -1), (1, 0), (0, 1), (-1, 0),
        (1, -1), (1, 1), (-1, 1), (-1, -1)
    ]
    return [(tile_x+o[0], tile_y+o[1]) for o in offsets]


def adjacent(tile_coord):
    """ Return the coordinates adjacent to the input coordinate.

    Return coordinates of the neighbourhood
    of the input coordinate, in the following order:

    7 0 4
    3   1
    6 2 5
    """
    return nine_grid(tile_coord)[1:]


def orthogonal(tile_coord):
    """ Return the coordinates orthogonal to the input coordinate.

    Return coordinates orthogonal to the input coordinate,
    in the following order:

      0
    3   1
      2
    """
    return nine_grid(tile_coord)[1:5]


class TileBoundingBox(object):
    """ Bounding box helper for checking tiles by their coordinates

    Defines a bounding box in pixel coordinates that allows
    checking conditions and retrieving in-tile pixel bounds
    for individual tiles, based on their coordinates.
    """

    def __init__(self, bbox):
        """
        Create a new TileBoundingBox based on a pixel bounding box.
        :param bbox: x, y, w, h bounding box in model pixel units
        :type bbox: lib.helpers.Rect
        """
        bbx, bby, bbw, bbh = bbox
        bb_rx, bb_ry = bbx + bbw - 1, bby + bbh - 1

        self.min_tx = int(bbx // N)
        self.min_ty = int(bby // N)
        self.max_tx = int(bb_rx // N)
        self.max_ty = int(bb_ry // N)

        self.min_px = int(bbx % N)
        self.min_py = int(bby % N)
        self.max_px = int(bb_rx % N)
        self.max_py = int(bb_ry % N)
        self.no_tile_crossing = (
            (self.min_px, self.min_py, self.max_px, self.max_py) ==
            (0, 0, N - 1, N - 1)
        )

    def tile_bounds(self, tc):
        """ Return the in-tile pixel bounds as a 4-tuple.
        Bounds cover the entire tile, unless it crosses
        an edge of the bounding box. Does not check if
        the tile actually lies inside the bounding box.
        """
        if self.no_tile_crossing:
            return 0, 0, N - 1, N - 1
        tx, ty = tc
        min_x = self.min_px if tx == self.min_tx else 0
        min_y = self.min_py if ty == self.min_ty else 0
        max_x = self.max_px if tx == self.max_tx else N - 1
        max_y = self.max_py if ty == self.max_ty else N - 1
        return min_x, min_y, max_x, max_y

    def outside(self, tc):
        """ Check if tile is outside bounding box.
        Checks if the tile of the given coordinate
        lies completely outside of the bounding box.
        """
        tx, ty = tc
        return (
            tx < self.min_tx or tx > self.max_tx or
            ty < self.min_ty or ty > self.max_ty
        )

    def crossing(self, tc):
        """ Check if tile crosses the bounding box.
        Checks if the tile of the given coordinate
        crosses at least one edge of the bounding box.
        """
        if self.no_tile_crossing:
            return False
        tx, ty = tc
        return (
            (tx == self.min_tx and self.min_px != 0) or
            (ty == self.min_ty and self.min_py != 0) or
            (tx == self.max_tx and self.max_px != (N - 1)) or
            (ty == self.max_ty and self.max_py != (N - 1))
        )

    def inside(self, tc):
        """ Check if tile is inside the bounding box.
        Checks if the tile of the given coordinate
        is fully enclosed by the bounding box.
        """
        tx, ty = tc
        if self.crossing(tc):
            return False
        return (
            self.min_tx <= tx <= self.max_tx and
            self.min_ty <= ty <= self.max_ty
        )



# Copyright (C) 2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Specifies gi module requirements & imports the binding modules

The GObject Introspection (gi) bindings should always be imported via this
module, and never directly from ``gi.repository``. This is to make sure that
the correct versions are always specified before import, even when individual
modules are being loaded in isolation (for purposes of testing).
"""

from __future__ import division, print_function

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("GdkPixbuf", "2.0")
gi.require_version("Gio", "2.0")
gi.require_version("GLib", "2.0")
gi.require_version("GObject", "2.0")
gi.require_version("Gtk", "3.0")
gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")

from gi.repository import Gdk  # noqa
from gi.repository import GdkPixbuf  # noqa
from gi.repository import Gio  # noqa
from gi.repository import GLib  # noqa
from gi.repository import GObject  # noqa
from gi.repository import Pango  # noqa
from gi.repository import PangoCairo  # noqa


# The import of the actual Gtk bindings needs to be deferred until the locale
# has been configured, in order for locale-specific layouts to be respected
# (right-to-left layouts).
# A GtkWrapper instance is used as a go-between that only triggers the real
# import when an attribute of the module is requested.

class GtkWrapper(object):

    def __getattribute__(self, attr):
        from gi.repository import Gtk as RealGtk
        return getattr(RealGtk, attr)

    def __getattr__(self, attr):
        return self.__getattribute__(attr)


Gtk = GtkWrapper()
